import numpy as np
import pytest
import scipy
import scipy.signal

from audiomentations import (
    BandPassFilter,
    BandStopFilter,
    HighPassFilter,
    LowPassFilter,
    PeakingFilter,
    LowShelfFilter,
    HighShelfFilter,
)

DEBUG = False


def get_chirp_test(sample_rate, duration):
    """Create a `duration` seconds chirp from 0Hz to `Nyquist frequency`"""
    n = np.arange(0, duration, 1 / sample_rate)
    samples = scipy.signal.chirp(n, 0, duration, sample_rate // 2, method="linear")
    return samples.astype(np.float32)


def get_randn_test(sample_rate, duration):
    """Create a random noise test stimulus"""
    n_samples = int(duration * sample_rate)
    samples = np.random.randn(n_samples)
    return samples.astype(np.float32)


class TestPeakingFilterTransforms:
    @pytest.mark.parametrize("center_freq", [4000.0])
    @pytest.mark.parametrize("gain_db", [-6.0, +6.0])
    @pytest.mark.parametrize("q_factor", [1.0])
    def test_one_single_input(self, center_freq, gain_db, q_factor):
        np.random.seed(1)

        sample_rate = 16000

        # Parameters for computing periodograms.
        # When examining lower frequencies we need to have
        # a high nfft number.

        nfft = 1024
        nperseg = 1024

        samples = get_chirp_test(sample_rate, 40)

        augment = PeakingFilter(
            min_center_freq=center_freq,
            max_center_freq=center_freq,
            min_gain_db=gain_db,
            max_gain_db=gain_db,
            min_q=q_factor,
            max_q=q_factor,
            p=1.0,
        )

        processed_samples = augment(samples, sample_rate=sample_rate)

        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

        # Compute periodograms
        wx, samples_pxx = scipy.signal.welch(
            samples,
            fs=sample_rate,
            nfft=nfft,
            nperseg=nperseg,
            scaling="spectrum",
            window="hann",
        )
        _, processed_samples_pxx = scipy.signal.welch(
            processed_samples,
            fs=sample_rate,
            nperseg=nperseg,
            nfft=nfft,
            scaling="spectrum",
            window="hann",
        )

        center_freq = augment.parameters["center_freq"]

        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(f"Filter type: Peaking at {gain_db} with Q={q_factor}")
            plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
            plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
            plt.legend(
                [
                    "Input signal",
                    f"Peaking with center freq {augment.parameters['center_freq']:.2f}",
                ]
            )
            plt.axvline(augment.parameters["center_freq"], color="red", linestyle=":")

            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Magnitude (dB)")
            plt.show()

        frequencies_of_interest = np.array([0.0, center_freq, sample_rate / 2])
        expected_differences = np.array([0.0, -gain_db, 0.0])

        for n, freq in enumerate(frequencies_of_interest):
            input_value_db = 10 * np.log10(
                samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            output_value_db = 10 * np.log10(
                processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            assert np.isclose(
                input_value_db - output_value_db,
                expected_differences[n],
                atol=1.0,
            )

    @pytest.mark.parametrize("center_freq", [10.0, 2000.0, 3900.0])
    @pytest.mark.parametrize("gain_db", [0.0, -6.0, +6.0])
    @pytest.mark.parametrize("q_factor", [0.1, 1.0, 10.0])
    @pytest.mark.parametrize("num_channels", [1, 2, 3])
    def test_multi_channel_input(self, center_freq, gain_db, q_factor, num_channels):

        sample_rate = 8000
        samples = get_randn_test(sample_rate, 10)

        # Convert to 2D N channels
        n_channels = np.tile(samples, (num_channels, 1))

        augment = PeakingFilter(
            min_center_freq=center_freq,
            max_center_freq=center_freq,
            min_gain_db=gain_db,
            max_gain_db=gain_db,
            min_q=q_factor,
            max_q=q_factor,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

        assert processed_n_channels.shape[0] == num_channels
        assert processed_n_channels.shape == n_channels.shape
        assert processed_n_channels.dtype == np.float32

        # Check that the processed 2D channel version applies the same effect
        # as the passband version.
        for _, channel in enumerate(processed_n_channels):
            if DEBUG:
                import matplotlib.pyplot as plt

                plt.title(f"Filter type: Peaking at {gain_db} with Q={q_factor}")

                plt.plot(processed_samples)
                plt.plot(channel, "r--")

                plt.legend(["1D", "2D"])
                plt.show()
            assert np.allclose(channel, processed_samples)


class TestLowShelfFilterTransform:
    @pytest.mark.parametrize("center_freq", [2000.0, 3900.0])
    @pytest.mark.parametrize("gain_db", [-6.0, +6.0, 0.0])
    @pytest.mark.parametrize("q_factor", [0.1, 1.0])
    def test_one_single_input(self, center_freq, gain_db, q_factor):
        np.random.seed(1)

        sample_rate = 16000

        # Parameters for computing periodograms.
        # When examining lower frequencies we need to have
        # a high nfft number.

        nfft = 4096
        nperseg = 1024

        samples = get_chirp_test(sample_rate, 40)

        augment = LowShelfFilter(
            min_center_freq=center_freq,
            max_center_freq=center_freq,
            min_gain_db=gain_db,
            max_gain_db=gain_db,
            min_q=q_factor,
            max_q=q_factor,
            p=1.0,
        )

        processed_samples = augment(samples, sample_rate=sample_rate)

        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

        # Compute periodograms
        wx, samples_pxx = scipy.signal.welch(
            samples,
            fs=sample_rate,
            nfft=nfft,
            nperseg=nperseg,
            scaling="spectrum",
            window="hann",
        )
        _, processed_samples_pxx = scipy.signal.welch(
            processed_samples,
            fs=sample_rate,
            nperseg=nperseg,
            nfft=nfft,
            scaling="spectrum",
            window="hann",
        )

        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(f"Filter type: High-shelf at {gain_db} with Q={q_factor}")
            plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
            plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
            plt.legend(
                [
                    "Input signal",
                    f"Peaking with center freq {augment.parameters['center_freq']:.2f}",
                ]
            )
            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Magnitude (dB)")
            plt.show()

        frequencies_of_interest = np.array([0.0, center_freq, sample_rate / 2])
        expected_differences = np.array([-gain_db, -gain_db / 2, 0.0])

        for n, freq in enumerate(frequencies_of_interest):
            input_value_db = 10 * np.log10(
                samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            output_value_db = 10 * np.log10(
                processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            assert np.isclose(
                input_value_db - output_value_db,
                expected_differences[n],
                atol=1.0,
            )

    @pytest.mark.parametrize("center_freq", [10.0, 2000.0, 3900.0])
    @pytest.mark.parametrize("gain_db", [0.0, -6.0, +6.0])
    @pytest.mark.parametrize("q_factor", [0.1, 1.0])
    @pytest.mark.parametrize("num_channels", [1, 2, 3])
    def test_n_channel_input(self, center_freq, gain_db, q_factor, num_channels):

        sample_rate = 8000
        samples = get_randn_test(sample_rate, 10)

        # Convert to 2D N channels
        n_channels = np.tile(samples, (num_channels, 1))

        augment = LowShelfFilter(
            min_center_freq=center_freq,
            max_center_freq=center_freq,
            min_gain_db=gain_db,
            max_gain_db=gain_db,
            min_q=q_factor,
            max_q=q_factor,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

        assert processed_n_channels.shape[0] == num_channels
        assert processed_n_channels.shape == n_channels.shape
        assert processed_n_channels.dtype == np.float32

        # Check that the processed 2D channel version applies the same effect
        # as the passband version.
        for _, channel in enumerate(processed_n_channels):
            if DEBUG:
                import matplotlib.pyplot as plt

                plt.title(f"Filter type: Low-shelf at {gain_db} with Q={q_factor}")

                plt.plot(processed_samples)
                plt.plot(channel, "r--")

                plt.legend(["1D", "2D"])
                plt.show()
            assert np.allclose(channel, processed_samples)


class TestHighShelfFilterTransform:
    @pytest.mark.parametrize("center_freq", [2000.0, 3900.0])
    @pytest.mark.parametrize("gain_db", [0.0, -6.0, +6.0])
    @pytest.mark.parametrize("q_factor", [0.1, 1.0])
    def test_one_single_input(self, center_freq, gain_db, q_factor):
        np.random.seed(1)

        sample_rate = 16000

        # Parameters for computing periodograms.
        # When examining lower frequencies we need to have
        # a high nfft number.

        nfft = 4096
        nperseg = 1024

        samples = get_chirp_test(sample_rate, 40)

        augment = HighShelfFilter(
            min_center_freq=center_freq,
            max_center_freq=center_freq,
            min_gain_db=gain_db,
            max_gain_db=gain_db,
            min_q=q_factor,
            max_q=q_factor,
            p=1.0,
        )

        processed_samples = augment(samples, sample_rate=sample_rate)

        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

        # Compute periodograms
        wx, samples_pxx = scipy.signal.welch(
            samples,
            fs=sample_rate,
            nfft=nfft,
            nperseg=nperseg,
            scaling="spectrum",
            window="hann",
        )
        _, processed_samples_pxx = scipy.signal.welch(
            processed_samples,
            fs=sample_rate,
            nperseg=nperseg,
            nfft=nfft,
            scaling="spectrum",
            window="hann",
        )

        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(f"Filter type: High-shelf at {gain_db} with Q={q_factor}")
            plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
            plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
            plt.legend(
                [
                    "Input signal",
                    f"Peaking with center freq {augment.parameters['center_freq']:.2f}",
                ]
            )
            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Magnitude (dB)")
            plt.show()

        frequencies_of_interest = np.array([0.0, center_freq, sample_rate / 2])
        expected_differences = np.array([0, -gain_db / 2, -gain_db])

        for n, freq in enumerate(frequencies_of_interest):
            input_value_db = 10 * np.log10(
                samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            output_value_db = 10 * np.log10(
                processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            assert np.isclose(
                input_value_db - output_value_db,
                expected_differences[n],
                atol=1.0,
            )

    @pytest.mark.parametrize("center_freq", [10.0, 2000.0, 3900.0])
    @pytest.mark.parametrize("gain_db", [0.0, -6.0, +6.0])
    @pytest.mark.parametrize("q_factor", [0.1, 1.0])
    @pytest.mark.parametrize("num_channels", [1, 2, 3])
    def test_n_channel_input(self, center_freq, gain_db, q_factor, num_channels):

        sample_rate = 8000
        samples = get_randn_test(sample_rate, 10)

        # Convert to 2D N channels
        n_channels = np.tile(samples, (num_channels, 1))

        augment = HighShelfFilter(
            min_center_freq=center_freq,
            max_center_freq=center_freq,
            min_gain_db=gain_db,
            max_gain_db=gain_db,
            min_q=q_factor,
            max_q=q_factor,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

        assert processed_n_channels.shape[0] == num_channels
        assert processed_n_channels.shape == n_channels.shape
        assert processed_n_channels.dtype == np.float32

        # Check that the processed 2D channel version applies the same effect
        # as the passband version.
        for _, channel in enumerate(processed_n_channels):
            if DEBUG:
                import matplotlib.pyplot as plt

                plt.title(f"Filter type: Low-shelf at {gain_db} with Q={q_factor}")

                plt.plot(processed_samples)
                plt.plot(channel, "r--")

                plt.legend(["1D", "2D"])
                plt.show()
            assert np.allclose(channel, processed_samples)


class TestLowPassFilterTransform:
    @pytest.mark.parametrize("cutoff_frequency", [1000])
    @pytest.mark.parametrize("rolloff", [6, 24])
    @pytest.mark.parametrize("zero_phase", [False, True])
    def test_one_single_input(
        self,
        cutoff_frequency,
        rolloff,
        zero_phase,
    ):

        sample_rate = 8000

        # Parameters for computing periodograms.
        # When examining lower frequencies we need to have
        # a high nfft number.

        nfft = 2048 * 2
        nperseg = 128

        samples = get_chirp_test(sample_rate, 10)

        # Expected dB drop at fc
        if zero_phase:
            expected_db_drop = 6
        else:
            expected_db_drop = 3

        if zero_phase and rolloff % 12 != 0:
            with pytest.raises(AssertionError):
                augment = LowPassFilter(
                    min_cutoff_freq=cutoff_frequency,
                    max_cutoff_freq=cutoff_frequency,
                    min_rolloff=rolloff,
                    max_rolloff=rolloff,
                    zero_phase=zero_phase,
                    p=1.0,
                )
            return
        else:
            augment = LowPassFilter(
                min_cutoff_freq=cutoff_frequency,
                max_cutoff_freq=cutoff_frequency,
                min_rolloff=rolloff,
                max_rolloff=rolloff,
                zero_phase=zero_phase,
                p=1.0,
            )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        # Compute periodograms
        wx, samples_pxx = scipy.signal.welch(
            samples,
            fs=sample_rate,
            nfft=nfft,
            nperseg=nperseg,
            scaling="spectrum",
            window="hann",
        )
        _, processed_samples_pxx = scipy.signal.welch(
            processed_samples,
            fs=sample_rate,
            nperseg=nperseg,
            nfft=nfft,
            scaling="spectrum",
            window="hann",
        )

        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(
                f"Filter type: High Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
            )
            plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
            plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
            plt.legend(["Input signal", f"Lowpassed at f_c={cutoff_frequency:.2f}"])
            plt.axvline(cutoff_frequency, color="red", linestyle=":")
            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Magnitude (dB)")
            plt.show()

        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

        frequencies_of_interest = np.array([0, cutoff_frequency, cutoff_frequency * 2])
        expected_differences = np.array(
            [0.0, expected_db_drop, expected_db_drop + rolloff]
        )

        # Tolerances for the differences in dB
        tolerances = np.array([1.0, 2.0, 5.0])

        for n, freq in enumerate(frequencies_of_interest):
            input_value_db = 10 * np.log10(
                samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            output_value_db = 10 * np.log10(
                processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            assert np.isclose(
                input_value_db - output_value_db,
                expected_differences[n],
                atol=tolerances[n],
            )

    @pytest.mark.parametrize("samples", [get_chirp_test(8000, 40)])
    @pytest.mark.parametrize("rolloff", [12, 120])
    @pytest.mark.parametrize("zero_phase", [False, True])
    @pytest.mark.parametrize("num_channels", [1, 2, 3])
    def test_n_channel_input(self, samples, rolloff, zero_phase, num_channels):

        sample_rate = 8000
        samples = get_randn_test(sample_rate, 10)

        # Convert to 2D N channels
        n_channels = np.tile(samples, (num_channels, 1))

        augment = LowPassFilter(
            min_cutoff_freq=1000,
            max_cutoff_freq=1000,
            min_rolloff=rolloff,
            max_rolloff=rolloff,
            zero_phase=zero_phase,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

        assert processed_n_channels.shape[0] == num_channels
        assert processed_n_channels.shape == n_channels.shape
        assert processed_n_channels.dtype == np.float32

        # Check that the processed 2D channel version applies the same effect
        # as the passband version.
        for _, channel in enumerate(processed_n_channels):
            if DEBUG:
                import matplotlib.pyplot as plt

                plt.title(
                    f"Filter type: Low-pass Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
                )
                plt.plot(processed_samples)
                plt.plot(channel, "r--")

                plt.legend(["1D", "2D"])
                plt.show()
            assert np.allclose(channel, processed_samples)


class TestHighPassFilterTransform:
    @pytest.mark.parametrize("cutoff_frequency", [1000])
    @pytest.mark.parametrize("rolloff", [6, 24])
    @pytest.mark.parametrize("zero_phase", [False, True])
    def test_one_single_input(self, cutoff_frequency, rolloff, zero_phase):

        sample_rate = 8000

        # Parameters for computing periodograms.
        # When examining lower frequencies we need to have
        # a high nfft number.

        nfft = 2048 * 2
        nperseg = 128

        samples = get_chirp_test(sample_rate, 10)

        # Expected dB drop at fc
        if zero_phase:
            expected_db_drop = 6
        else:
            expected_db_drop = 3

        if zero_phase and rolloff % 12 != 0:
            with pytest.raises(AssertionError):
                augment = HighPassFilter(
                    min_cutoff_freq=cutoff_frequency,
                    max_cutoff_freq=cutoff_frequency,
                    min_rolloff=rolloff,
                    max_rolloff=rolloff,
                    zero_phase=zero_phase,
                    p=1.0,
                )
            return
        else:
            augment = HighPassFilter(
                min_cutoff_freq=cutoff_frequency,
                max_cutoff_freq=cutoff_frequency,
                min_rolloff=rolloff,
                max_rolloff=rolloff,
                zero_phase=zero_phase,
                p=1.0,
            )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        # Compute periodograms
        wx, samples_pxx = scipy.signal.welch(
            samples,
            fs=sample_rate,
            nfft=nfft,
            nperseg=nperseg,
            scaling="spectrum",
            window="hann",
        )
        _, processed_samples_pxx = scipy.signal.welch(
            processed_samples,
            fs=sample_rate,
            nperseg=nperseg,
            nfft=nfft,
            scaling="spectrum",
            window="hann",
        )

        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(
                f"Filter type: High Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
            )
            plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
            plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
            plt.legend(["Input signal", f"Highpassed at f_c={cutoff_frequency:.2f}"])
            plt.axvline(cutoff_frequency, color="red", linestyle=":")
            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Magnitude (dB)")
            plt.show()

        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

        frequencies_of_interest = np.array(
            [cutoff_frequency / 2, cutoff_frequency, sample_rate / 2]
        )
        expected_differences = np.array(
            [expected_db_drop + rolloff, expected_db_drop, 0.0]
        )

        # Tolerances for the differences in dB
        tolerances = np.array([5.0, 2.0, 1.0])

        for n, freq in enumerate(frequencies_of_interest):
            input_value_db = 10 * np.log10(
                samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            output_value_db = 10 * np.log10(
                processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            assert np.isclose(
                input_value_db - output_value_db,
                expected_differences[n],
                atol=tolerances[n],
            )

    @pytest.mark.parametrize("samples", [get_chirp_test(8000, 40)])
    @pytest.mark.parametrize("rolloff", [12, 120])
    @pytest.mark.parametrize("zero_phase", [False, True])
    @pytest.mark.parametrize("num_channels", [1, 2, 3])
    def test_n_channel_input(self, samples, rolloff, zero_phase, num_channels):

        sample_rate = 8000
        samples = get_randn_test(sample_rate, 10)

        # Convert to 2D N channels
        n_channels = np.tile(samples, (num_channels, 1))

        augment = HighPassFilter(
            min_cutoff_freq=1000,
            max_cutoff_freq=1000,
            min_rolloff=rolloff,
            max_rolloff=rolloff,
            zero_phase=zero_phase,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

        assert processed_n_channels.shape[0] == num_channels
        assert processed_n_channels.shape == n_channels.shape
        assert processed_n_channels.dtype == np.float32

        # Check that the processed 2D channel version applies the same effect
        # as the passband version.
        for _, channel in enumerate(processed_n_channels):
            if DEBUG:
                import matplotlib.pyplot as plt

                plt.title(
                    f"Filter type: Highpass Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
                )
                plt.plot(processed_samples)
                plt.plot(channel, "r--")

                plt.legend(["1D", "2D"])
                plt.show()
            assert np.allclose(channel, processed_samples)

    @pytest.mark.parametrize("cutoff_frequency", [5000])
    @pytest.mark.parametrize("rolloff", [6])
    @pytest.mark.parametrize("zero_phase", [False])
    def test_nyquist_limit(self, cutoff_frequency, rolloff, zero_phase):
        # Test that the filter doesn't raise an exception when
        # cutoff_frequency is greater than the Nyquist frequency

        sample_rate = 8000

        samples = get_chirp_test(sample_rate, 3)

        augment = HighPassFilter(
            min_cutoff_freq=cutoff_frequency,
            max_cutoff_freq=cutoff_frequency,
            min_rolloff=rolloff,
            max_rolloff=rolloff,
            zero_phase=zero_phase,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert processed_samples.shape == samples.shape


class TestBandPassFilterTransform:
    @pytest.mark.parametrize("center_frequency", [3000])
    @pytest.mark.parametrize("bandwidth_fraction", [0.666])
    @pytest.mark.parametrize("rolloff", [6, 24])
    @pytest.mark.parametrize("zero_phase", [False, True])
    def test_one_single_input(
        self, center_frequency, bandwidth_fraction, rolloff, zero_phase
    ):

        sample_rate = 16000

        # Parameters for computing periodograms.
        # When examining lower frequencies we need to have
        # a high nfft number.

        nfft = 2048 * 2
        nperseg = 128

        samples = get_chirp_test(sample_rate, 10)

        # Expected dB drop at fc
        if zero_phase:
            expected_db_drop = 6
        else:
            expected_db_drop = 3

        if zero_phase and rolloff % 12 != 0:
            with pytest.raises(AssertionError):
                augment = BandPassFilter(
                    min_center_freq=center_frequency,
                    max_center_freq=center_frequency,
                    min_bandwidth_fraction=bandwidth_fraction,
                    max_bandwidth_fraction=bandwidth_fraction,
                    min_rolloff=rolloff,
                    max_rolloff=rolloff,
                    zero_phase=zero_phase,
                    p=1.0,
                )
            return
        else:
            augment = BandPassFilter(
                min_center_freq=center_frequency,
                max_center_freq=center_frequency,
                min_bandwidth_fraction=bandwidth_fraction,
                max_bandwidth_fraction=bandwidth_fraction,
                min_rolloff=rolloff,
                max_rolloff=rolloff,
                zero_phase=zero_phase,
                p=1.0,
            )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        # Compute periodograms
        wx, samples_pxx = scipy.signal.welch(
            samples,
            fs=sample_rate,
            nfft=nfft,
            nperseg=nperseg,
            scaling="spectrum",
            window="hann",
        )
        _, processed_samples_pxx = scipy.signal.welch(
            processed_samples,
            fs=sample_rate,
            nperseg=nperseg,
            nfft=nfft,
            scaling="spectrum",
            window="hann",
        )

        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(
                f"Filter type: Band-pass Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
            )
            plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
            plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
            plt.legend(["Input signal", f"Highpassed at f_c={center_frequency:.2f}"])
            plt.axvline(center_frequency, color="red", linestyle=":")
            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Magnitude (dB)")
            plt.show()

        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

        bandwidth = bandwidth_fraction * center_frequency
        left_cutoff_freq = center_frequency - bandwidth / 2
        right_cutoff_freq = center_frequency + bandwidth / 2

        frequencies_of_interest = np.array(
            [
                left_cutoff_freq,
                center_frequency,
                right_cutoff_freq,
            ]
        )
        expected_differences = np.array(
            [
                expected_db_drop,
                0,
                expected_db_drop,
            ]
        )

        # Tolerances for the differences in dB
        tolerances = np.array([2.0, 1.0, 2.0])

        for n, freq in enumerate(frequencies_of_interest):
            input_value_db = 10 * np.log10(
                samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            output_value_db = 10 * np.log10(
                processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            assert np.isclose(
                input_value_db - output_value_db,
                expected_differences[n],
                atol=tolerances[n],
            )

    @pytest.mark.parametrize("center_frequency", [3000])
    @pytest.mark.parametrize("bandwidth_fraction", [0.666])
    @pytest.mark.parametrize("samples", [get_chirp_test(8000, 40)])
    @pytest.mark.parametrize("rolloff", [12, 120])
    @pytest.mark.parametrize("zero_phase", [False, True])
    @pytest.mark.parametrize("num_channels", [1, 2, 3])
    def test_two_channel_input(
        self,
        center_frequency,
        bandwidth_fraction,
        samples,
        rolloff,
        zero_phase,
        num_channels,
    ):

        sample_rate = 16000
        samples = get_randn_test(sample_rate, 10)

        # Convert to 2D N channels
        n_channels = np.tile(samples, (num_channels, 1))

        augment = BandPassFilter(
            min_center_freq=center_frequency,
            max_center_freq=center_frequency,
            min_bandwidth_fraction=bandwidth_fraction,
            max_bandwidth_fraction=bandwidth_fraction,
            min_rolloff=rolloff,
            max_rolloff=rolloff,
            zero_phase=zero_phase,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

        assert processed_n_channels.shape[0] == num_channels
        assert processed_n_channels.shape == n_channels.shape
        assert processed_n_channels.dtype == np.float32

        # Check that the processed 2D channel version applies the same effect
        # as the passband version.
        for _, channel in enumerate(processed_n_channels):
            if DEBUG:
                import matplotlib.pyplot as plt

                plt.title(
                    f"Filter type: Band-pass Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
                )
                plt.plot(processed_samples)
                plt.plot(channel, "r--")

                plt.legend(["1D", "2D"])
                plt.show()
            assert np.allclose(channel, processed_samples)

    @pytest.mark.parametrize("center_frequency", [7000])
    @pytest.mark.parametrize("bandwidth_fraction", [0.666])
    @pytest.mark.parametrize("rolloff", [6])
    @pytest.mark.parametrize("zero_phase", [False])
    def test_nyquist_limit(
        self, center_frequency, bandwidth_fraction, rolloff, zero_phase
    ):
        # Test that the filter doesn't raise an exception when
        # center_freq + bandwidth / 2 is greater than the Nyquist frequency

        sample_rate = 16000

        samples = get_chirp_test(sample_rate, 3)

        augment = BandPassFilter(
            min_center_freq=center_frequency,
            max_center_freq=center_frequency,
            min_bandwidth_fraction=bandwidth_fraction,
            max_bandwidth_fraction=bandwidth_fraction,
            min_rolloff=rolloff,
            max_rolloff=rolloff,
            zero_phase=zero_phase,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert processed_samples.shape == samples.shape

    def test_too_extreme_bandwidth_fractions(self):
        with pytest.raises(ValueError):
            BandPassFilter(
                min_bandwidth_fraction=0.0,
                max_bandwidth_fraction=1.0,
                p=1.0,
            )

        with pytest.raises(ValueError):
            BandPassFilter(
                min_bandwidth_fraction=0.1,
                max_bandwidth_fraction=3.0,
                p=1.0,
            )

        with pytest.raises(ValueError):
            BandPassFilter(
                min_bandwidth_fraction=0.5,
                max_bandwidth_fraction=0.1,
                p=1.0,
            )


class TestBandStopFilterTransform:
    @pytest.mark.parametrize("center_frequency", [3000])
    @pytest.mark.parametrize("bandwidth_fraction", [0.666])
    @pytest.mark.parametrize("rolloff", [6, 24])
    @pytest.mark.parametrize("zero_phase", [False, True])
    def test_one_single_input(
        self, center_frequency, bandwidth_fraction, rolloff, zero_phase
    ):

        sample_rate = 16000

        # Parameters for computing periodograms.
        # When examining lower frequencies we need to have
        # a high nfft number.

        nfft = 2048 * 2
        nperseg = 128

        samples = get_chirp_test(sample_rate, 10)

        # Expected dB drop at fc
        if zero_phase:
            expected_db_drop = 6
        else:
            expected_db_drop = 3

        if zero_phase and rolloff % 12 != 0:
            with pytest.raises(AssertionError):
                augment = BandStopFilter(
                    min_center_freq=center_frequency,
                    max_center_freq=center_frequency,
                    min_bandwidth_fraction=bandwidth_fraction,
                    max_bandwidth_fraction=bandwidth_fraction,
                    min_rolloff=rolloff,
                    max_rolloff=rolloff,
                    zero_phase=zero_phase,
                    p=1.0,
                )
            return
        else:
            augment = BandStopFilter(
                min_center_freq=center_frequency,
                max_center_freq=center_frequency,
                min_bandwidth_fraction=bandwidth_fraction,
                max_bandwidth_fraction=bandwidth_fraction,
                min_rolloff=rolloff,
                max_rolloff=rolloff,
                zero_phase=zero_phase,
                p=1.0,
            )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        # Compute periodograms
        wx, samples_pxx = scipy.signal.welch(
            samples,
            fs=sample_rate,
            nfft=nfft,
            nperseg=nperseg,
            scaling="spectrum",
            window="hann",
        )
        _, processed_samples_pxx = scipy.signal.welch(
            processed_samples,
            fs=sample_rate,
            nperseg=nperseg,
            nfft=nfft,
            scaling="spectrum",
            window="hann",
        )

        if DEBUG:
            import matplotlib.pyplot as plt

            plt.title(
                f"Filter type: Band-stop Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
            )
            plt.plot(wx, 10 * np.log10(np.abs(samples_pxx)))
            plt.plot(wx, 10 * np.log10(np.abs(processed_samples_pxx)), ":")
            plt.legend(["Input signal", f"Highpassed at f_c={center_frequency:.2f}"])
            plt.axvline(center_frequency, color="red", linestyle=":")
            plt.xlabel("Frequency (Hz)")
            plt.ylabel("Magnitude (dB)")
            plt.show()

        assert processed_samples.shape == samples.shape
        assert processed_samples.dtype == np.float32

        bandwidth = bandwidth_fraction * center_frequency
        left_cutoff_freq = center_frequency - bandwidth / 2
        right_cutoff_freq = center_frequency + bandwidth / 2

        frequencies_of_interest = np.array(
            [0, left_cutoff_freq, right_cutoff_freq, sample_rate / 2]
        )
        expected_differences = np.array([0, expected_db_drop, expected_db_drop, 0.0])

        # Tolerances for the differences in dB
        tolerances = np.array([1.0, 2.0, 2.0, 1.0])

        for n, freq in enumerate(frequencies_of_interest):
            input_value_db = 10 * np.log10(
                samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            output_value_db = 10 * np.log10(
                processed_samples_pxx[int(np.round(nfft / sample_rate * freq))]
            )

            assert np.isclose(
                input_value_db - output_value_db,
                expected_differences[n],
                atol=tolerances[n],
            )

    @pytest.mark.parametrize("center_frequency", [3000])
    @pytest.mark.parametrize("bandwidth_fraction", [0.666])
    @pytest.mark.parametrize("samples", [get_chirp_test(8000, 40)])
    @pytest.mark.parametrize("rolloff", [12, 120])
    @pytest.mark.parametrize("zero_phase", [False, True])
    @pytest.mark.parametrize("num_channels", [1, 2, 3])
    def test_two_channel_input(
        self,
        center_frequency,
        bandwidth_fraction,
        samples,
        rolloff,
        zero_phase,
        num_channels,
    ):

        sample_rate = 16000
        samples = get_randn_test(sample_rate, 10)

        # Convert to 2D N channels
        n_channels = np.tile(samples, (num_channels, 1))

        augment = BandStopFilter(
            min_center_freq=center_frequency,
            max_center_freq=center_frequency,
            min_bandwidth_fraction=bandwidth_fraction,
            max_bandwidth_fraction=bandwidth_fraction,
            min_rolloff=rolloff,
            max_rolloff=rolloff,
            zero_phase=zero_phase,
            p=1.0,
        )

        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        processed_n_channels = augment(samples=n_channels, sample_rate=sample_rate)

        assert processed_n_channels.shape[0] == num_channels
        assert processed_n_channels.shape == n_channels.shape
        assert processed_n_channels.dtype == np.float32

        # Check that the processed 2D channel version applies the same effect
        # as the passband version.
        for _, channel in enumerate(processed_n_channels):
            if DEBUG:
                import matplotlib.pyplot as plt

                plt.title(
                    f"Filter type: Band-stop Roll-off:{rolloff}dB/octave Zero-phase:{zero_phase}"
                )
                plt.plot(processed_samples)
                plt.plot(channel, "r--")

                plt.legend(["1D", "2D"])
                plt.show()
            assert np.allclose(channel, processed_samples)



import numpy as np

from audiomentations import TimeStretch


class TestTimeStretch:
    def test_dynamic_length(self):
        samples = np.zeros((2048,), dtype=np.float32)
        sample_rate = 16000
        augmenter = TimeStretch(
            min_rate=0.8, max_rate=0.9, leave_length_unchanged=False, p=1.0
        )

        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert len(samples) > 2048

    def test_fixed_length(self):
        samples = np.zeros((2048,), dtype=np.float32)
        sample_rate = 16000
        augmenter = TimeStretch(
            min_rate=0.8, max_rate=0.9, leave_length_unchanged=True, p=1.0
        )

        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert len(samples) == 2048

    def test_multichannel(self):
        num_channels = 3
        samples = np.random.normal(0, 0.1, size=(num_channels, 5555)).astype(np.float32)
        sample_rate = 16000
        augmenter = TimeStretch(
            min_rate=0.8, max_rate=0.9, leave_length_unchanged=True, p=1.0
        )

        samples_out = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == samples_out.dtype
        assert samples.shape == samples_out.shape
        for i in range(num_channels):
            assert not np.allclose(samples[i], samples_out[i])



import numpy as np
import pyloudnorm
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal

from audiomentations import Gain
from audiomentations.core.post_gain import PostGain
from audiomentations.core.utils import calculate_rms, get_max_abs_amplitude


class TestPostGain:
    def test_same_rms(self):
        samples = np.array([1.0, 0.5, -0.25, -0.125, 0.0], dtype=np.float32)
        sample_rate = 16000

        augment = PostGain(
            Gain(min_gain_db=-6, max_gain_db=-6, p=1.0), method="same_rms"
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert_almost_equal(
            calculate_rms(processed_samples),
            calculate_rms(samples),
        )
        assert processed_samples.dtype == np.float32

    def test_same_lufs(self):
        samples = np.random.uniform(low=-0.5, high=0.5, size=(2, 8000)).astype(
            np.float32
        )
        sample_rate = 16000

        augment = PostGain(
            Gain(min_gain_db=60, max_gain_db=60, p=1.0), method="same_lufs"
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        meter = pyloudnorm.Meter(sample_rate)  # create BS.1770 meter
        lufs_before = meter.integrated_loudness(samples.transpose())
        lufs_after = meter.integrated_loudness(processed_samples.transpose())
        assert_almost_equal(lufs_after, lufs_before, decimal=6)
        assert processed_samples.dtype == np.float32

    def test_peak_normalize_always(self):
        samples = np.random.uniform(low=-0.5, high=0.5, size=(2, 8000)).astype(
            np.float32
        )
        sample_rate = 16000

        augment = PostGain(
            Gain(min_gain_db=-55, max_gain_db=-55, p=1.0),
            method="peak_normalize_always",
        )
        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        assert get_max_abs_amplitude(processed_samples) == pytest.approx(1.0)
        assert processed_samples.dtype == np.float32

    def test_peak_normalize_if_too_loud(self):
        samples = np.array(
            [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        sample_rate = 16000
        augmenter = PostGain(
            Gain(min_gain_db=0.0, max_gain_db=0.0, p=1.0),
            method="peak_normalize_if_too_loud",
        )

        processed_samples = augmenter(samples=samples, sample_rate=sample_rate)
        assert_array_equal(processed_samples, samples)
        assert processed_samples.dtype == np.float32

        samples_too_loud = np.array(
            [[0.9, 0.5, -0.25, -0.125, 0.0], [1.2, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        processed_samples = augmenter(samples=samples_too_loud, sample_rate=sample_rate)
        assert_array_equal(processed_samples, samples_too_loud / 1.2)
        assert processed_samples.dtype == np.float32



import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class Reverse(BaseWaveformTransform):
    """
    Reverse the audio, also known as time inversion. Inversion of an audio track along its time axis is
    analogous to the random flip of an image, an augmentation technique widely used in the visual domain.
    This can be relevant in the context of audio classification. It was successfully applied in the paper
    AudioCLIP: Extending CLIP to Image, Text and Audio
    https://arxiv.org/pdf/2106.13043.pdf
    """

    supports_multichannel = True

    def __init__(self, p: float = 0.5):
        """
        :param p: The probability of applying this transform
        """
        super().__init__(p)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        if len(samples.shape) > 1:
            return np.fliplr(samples)
        else:
            return np.flipud(samples)



from audiomentations.augmentations.base_butterword_filter import BaseButterworthFilter


class HighPassFilter(BaseButterworthFilter):
    """
    Apply high-pass filtering to the input audio of parametrized filter steepness (6/12/18... dB / octave).
    Can also be set for zero-phase filtering (will result in a 6 dB drop at cutoff).
    """

    supports_multichannel = True

    def __init__(
        self,
        min_cutoff_freq: float = 20.0,
        max_cutoff_freq: float = 2400.0,
        min_rolloff: int = 12,
        max_rolloff: int = 24,
        zero_phase: bool = False,
        p: float = 0.5,
    ):
        """
        :param min_cutoff_freq: Minimum cutoff frequency in hertz
        :param max_cutoff_freq: Maximum cutoff frequency in hertz
        :param min_rolloff: Minimum filter roll-off (in dB/octave).
            Must be a multiple of 6
        :param max_rolloff: Maximum filter roll-off (in dB/octave)
            Must be a multiple of 6
        :param zero_phase: Whether filtering should be zero phase.
            When this is set to `True`, it will not affect the phase of the
            input signal but will sound 3 dB lower at the cutoff frequency
            compared to the non-zero phase case (6 dB vs. 3 dB). Additionally,
            it is twice as slow as the non-zero phase case. If you
            absolutely want no phase distortions (e.g. want to augment a
            drum track), set this to `True`.
        :param p: The probability of applying this transform
        """
        super().__init__(
            min_cutoff_freq=min_cutoff_freq,
            max_cutoff_freq=max_cutoff_freq,
            min_rolloff=min_rolloff,
            max_rolloff=max_rolloff,
            zero_phase=zero_phase,
            p=p,
            filter_type="highpass",
        )



import random

import librosa
import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class Resample(BaseWaveformTransform):
    """
    Resample signal using librosa.core.resample

    To do downsampling only, set both minimum and maximum sampling rate lower than
    original sampling rate. Conversely, to perform upsampling only, set both rates higher than the original sampling
    rate.
    """

    supports_multichannel = True

    def __init__(
        self, min_sample_rate: int = 8000, max_sample_rate: int = 44100, p: float = 0.5
    ):
        """
        :param min_sample_rate: The minimum sample rate
        :param max_sample_rate: The maximum sample rate
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        assert min_sample_rate <= max_sample_rate
        self.min_sample_rate = min_sample_rate
        self.max_sample_rate = max_sample_rate

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["target_sample_rate"] = random.randint(
                self.min_sample_rate, self.max_sample_rate
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        samples = librosa.core.resample(
            samples,
            orig_sr=sample_rate,
            target_sr=self.parameters["target_sample_rate"],
        )
        return samples



import numpy as np
import pytest

from audiomentations import BitCrush


class TestBitCrush:
    def test_single_channel(self):
        samples = np.random.normal(0, 0.1, size=(2048,)).astype(np.float32)
        sample_rate = 16000
        augmenter = BitCrush(min_bit_depth=3, max_bit_depth=6, p=1.0)

        distorted_samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == distorted_samples.dtype
        assert samples.shape == distorted_samples.shape
        assert 2 ** augmenter.parameters["bit_depth"] + 1 > len(
            np.unique(np.round(distorted_samples, 5))
        )

    def test_multichannel(self):
        num_channels = 3
        samples = np.random.normal(0, 0.1, size=(num_channels, 2048)).astype(np.float32)
        sample_rate = 16000
        augmenter = BitCrush(min_bit_depth=3, max_bit_depth=6, p=1.0)

        distorted_samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == distorted_samples.dtype
        assert samples.shape == distorted_samples.shape
        assert 2 ** augmenter.parameters["bit_depth"] + 1 > len(
            np.unique(np.round(distorted_samples, 5))
        )

    def test_param_range(self):
        with pytest.raises(ValueError):
            BitCrush(min_bit_depth=0, max_bit_depth=6, p=1.0)
        with pytest.raises(ValueError):
            BitCrush(min_bit_depth=7, max_bit_depth=64, p=1.0)
        with pytest.raises(ValueError):
            BitCrush(min_bit_depth=8, max_bit_depth=6, p=1.0)



import functools
import random
import warnings
from pathlib import Path
from typing import Optional, List, Callable, Union

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.audio_loading_utils import load_sound_file
from audiomentations.core.transforms_interface import BaseWaveformTransform
from audiomentations.core.utils import (
    calculate_desired_noise_rms,
    calculate_rms,
    convert_decibels_to_amplitude_ratio,
    find_audio_files_in_paths,
)


class AddBackgroundNoise(BaseWaveformTransform):
    """Mix in another sound, e.g. a background noise. Useful if your original sound is clean and
    you want to simulate an environment where background noise is present.
    Can also be used for mixup, as in https://arxiv.org/pdf/1710.09412.pdf
    A folder of (background noise) sounds to be mixed in must be specified. These sounds should
    ideally be at least as long as the input sounds to be transformed. Otherwise, the background
    sound will be repeated, which may sound unnatural.
    Note that the gain of the added noise is relative to the signal level in the input if the parameter noise_rms
    is set to "relative" (default option). This implies that if the input is completely silent, no noise will be added.
    Here are some examples of datasets that can be downloaded and used as background noise:
    * https://github.com/karolpiczak/ESC-50#download
    * https://github.com/microsoft/DNS-Challenge/
    """

    def __init__(
        self,
        sounds_path: Union[List[Path], List[str], Path, str],
        min_snr_db: float = 3.0,
        max_snr_db: float = 30.0,
        noise_rms: str = "relative",
        min_absolute_rms_db: float = -45.0,
        max_absolute_rms_db: float = -15.0,
        noise_transform: Optional[
            Callable[[NDArray[np.float32], int], NDArray[np.float32]]
        ] = None,
        p: float = 0.5,
        lru_cache_size: int = 2,
    ):
        """
        :param sounds_path: A path or list of paths to audio file(s) and/or folder(s) with
            audio files. Can be str or Path instance(s). The audio files given here are
            supposed to be background noises.
        :param min_snr_db: Minimum signal-to-noise ratio in dB. Is only used if noise_rms is set to "relative"
        :param max_snr_db: Maximum signal-to-noise ratio in dB. Is only used if noise_rms is set to "relative"
        :param noise_rms: Defines how the background noise will be added to the audio input. If the chosen
            option is "relative", the RMS of the added noise will be proportional to the RMS of
            the input sound. If the chosen option is "absolute", the background noise will have
            an RMS independent of the RMS of the input audio file. The default option is "relative".
        :param min_absolute_rms_db: Is only used if noise_rms is set to "absolute". It is
            the minimum RMS value in dB that the added noise can take. The lower the RMS is,
            the lower the added sound will be. Default: -45.0
        :param max_absolute_rms_db: Is only used if noise_rms is set to "absolute". It is
            the maximum RMS value in dB that the added noise can take. Note that this value
            can not exceed 0. Default: -15.0
        :param noise_transform: A callable waveform transform (or composition of transforms) that
            gets applied to the noise before it gets mixed in. The callable is expected
            to input audio waveform (numpy array) and sample rate (int).
        :param p: The probability of applying this transform
        :param lru_cache_size: Maximum size of the LRU cache for storing noise files in memory
        """
        super().__init__(p)
        self.sounds_path = sounds_path
        self.sound_file_paths = find_audio_files_in_paths(self.sounds_path)
        self.sound_file_paths = [str(p) for p in self.sound_file_paths]

        assert len(self.sound_file_paths) > 0

        if min_snr_db > max_snr_db:
            raise ValueError("min_snr_db must not be greater than max_snr_db")
        self.min_snr_db = min_snr_db
        self.max_snr_db = max_snr_db

        if min_absolute_rms_db > max_absolute_rms_db:
            raise ValueError(
                "min_absolute_rms_db must not be greater than max_absolute_rms_db"
            )
        if max_absolute_rms_db > 0:
            raise ValueError("max_absolute_rms_db must not be greater than 0")
        self.min_absolute_rms_db = min_absolute_rms_db
        self.max_absolute_rms_db = max_absolute_rms_db

        self.noise_rms = noise_rms
        self.lru_cache_size = lru_cache_size
        self._load_sound = functools.lru_cache(maxsize=self.lru_cache_size)(
            AddBackgroundNoise._load_sound
        )
        self.noise_transform = noise_transform

    @staticmethod
    def _load_sound(file_path, sample_rate):
        return load_sound_file(file_path, sample_rate)

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["snr_db"] = random.uniform(self.min_snr_db, self.max_snr_db)
            self.parameters["rms_db"] = random.uniform(
                self.min_absolute_rms_db, self.max_absolute_rms_db
            )
            self.parameters["noise_file_path"] = random.choice(self.sound_file_paths)

            num_samples = len(samples)
            noise_sound, _ = self._load_sound(
                self.parameters["noise_file_path"], sample_rate
            )

            num_noise_samples = len(noise_sound)
            min_noise_offset = 0
            max_noise_offset = max(0, num_noise_samples - num_samples - 1)
            self.parameters["noise_start_index"] = random.randint(
                min_noise_offset, max_noise_offset
            )
            self.parameters["noise_end_index"] = (
                self.parameters["noise_start_index"] + num_samples
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        noise_sound, _ = self._load_sound(
            self.parameters["noise_file_path"], sample_rate
        )
        noise_sound = noise_sound[
            self.parameters["noise_start_index"] : self.parameters["noise_end_index"]
        ]

        if self.noise_transform:
            noise_sound = self.noise_transform(noise_sound, sample_rate)

        noise_rms = calculate_rms(noise_sound)
        if noise_rms < 1e-9:
            warnings.warn(
                "The file {} is too silent to be added as noise. Returning the input"
                " unchanged.".format(self.parameters["noise_file_path"])
            )
            return samples

        clean_rms = calculate_rms(samples)

        if self.noise_rms == "relative":
            desired_noise_rms = calculate_desired_noise_rms(
                clean_rms, self.parameters["snr_db"]
            )

            # Adjust the noise to match the desired noise RMS
            noise_sound = noise_sound * (desired_noise_rms / noise_rms)

        if self.noise_rms == "absolute":
            desired_noise_rms_db = self.parameters["rms_db"]
            desired_noise_rms_amp = convert_decibels_to_amplitude_ratio(
                desired_noise_rms_db
            )
            gain = desired_noise_rms_amp / noise_rms
            noise_sound = noise_sound * gain

        # Repeat the sound if it shorter than the input sound
        num_samples = len(samples)
        while len(noise_sound) < num_samples:
            noise_sound = np.concatenate((noise_sound, noise_sound))

        if len(noise_sound) > num_samples:
            noise_sound = noise_sound[0:num_samples]

        # Return a mix of the input sound and the background noise sound
        return samples + noise_sound

    def __getstate__(self):
        state = self.__dict__.copy()
        warnings.warn(
            "Warning: the LRU cache of AddBackgroundNoise gets discarded when pickling"
            " it. E.g. this means the cache will not be used when using"
            " AddBackgroundNoise together with multiprocessing on Windows"
        )
        del state["_load_sound"]
        return state



import json
import os
import pickle
import random

import numpy as np
import pytest
from numpy.testing import assert_array_equal, assert_array_almost_equal

from audiomentations import AddShortNoises, PolarityInversion
from audiomentations.core.transforms_interface import (
    MultichannelAudioNotSupportedException,
)
from audiomentations.core.utils import calculate_rms
from demo.demo import DEMO_DIR


class TestAddShortNoises:
    def test_add_short_noises(self):
        sample_rate = 44100
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
            np.float32
        )
        rms_before = calculate_rms(samples)
        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_time_between_sounds=2.0,
            max_time_between_sounds=8.0,
            p=1.0,
        )
        samples_out = augmenter(samples=samples, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples.shape

        rms_after = calculate_rms(samples_out)
        assert rms_after > rms_before

    def test_add_short_noises_with_signal_gain_during_noise(self):
        sample_rate = 44100
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
            np.float32
        )
        rms_before = calculate_rms(samples)
        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_snr_db=50.0,
            max_snr_db=50.0,
            min_time_between_sounds=2.0,
            max_time_between_sounds=4.0,
            signal_gain_db_during_noise=-100,
            p=1.0,
        )
        samples_out = augmenter(samples=samples, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples.shape

        rms_after = calculate_rms(samples_out)
        assert rms_after < rms_before

    def test_add_short_noises_with_noise_transform(self):
        sample_rate = 44100
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
            np.float32
        )
        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_snr_db=50.0,
            max_snr_db=50.0,
            min_time_between_sounds=1.0,
            max_time_between_sounds=2.0,
            noise_transform=PolarityInversion(p=1.0),
            p=1.0,
        )
        samples_out_with_noise_transform = augmenter(
            samples=samples, sample_rate=sample_rate
        )
        assert samples_out_with_noise_transform.dtype == np.float32
        assert samples_out_with_noise_transform.shape == samples.shape

        augmenter.noise_transform = None
        samples_out_without_noise_transform = augmenter(
            samples=samples, sample_rate=sample_rate
        )

        with np.testing.assert_raises(AssertionError):
            assert_array_almost_equal(
                samples_out_without_noise_transform, samples_out_with_noise_transform
            )

    def test_input_shorter_than_noise(self):
        """
        Verify correct behavior when the input sound is shorter than the added noise sounds.
        """
        sample_rate = 44100
        samples = np.sin(
            np.linspace(0, 440 * 2 * np.pi, int(0.03125 * sample_rate))
        ).astype(np.float32)
        rms_before = calculate_rms(samples)
        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_time_between_sounds=0.00001,
            max_time_between_sounds=0.00002,
            p=1.0,
        )
        samples_out = augmenter(samples=samples, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples.shape

        rms_after = calculate_rms(samples_out)
        assert rms_after > rms_before

    def test_add_silence(self):
        """Check that AddShortNoises does not crash if a noise is completely silent."""
        sample_rate = 48000
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 7000)).astype(np.float32)
        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "silence"),
            min_time_between_sounds=0.001,
            max_time_between_sounds=0.002,
            p=1.0,
        )

        augmenter.parameters = {
            "should_apply": True,
            "sounds": [
                {
                    "fade_in_time": 0.04257633246838298,
                    "start": -0.00013191289693534575,
                    "end": 0.8071696744046519,
                    "fade_out_time": 0.07119110196424423,
                    "file_path": os.path.join(DEMO_DIR, "silence", "silence.wav"),
                    "snr_db": 19.040001423519563,
                }
            ],
        }
        augmenter.freeze_parameters()

        samples_out = augmenter(samples=samples, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples.shape

    def test_too_long_fade_time(self):
        """Check that a too long fade time does not result in an exception."""
        sample_rate = 44100
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
            np.float32
        )
        rms_before = calculate_rms(samples)
        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_time_between_sounds=2.0,
            max_time_between_sounds=8.0,
            min_fade_in_time=0.9,
            max_fade_in_time=0.99,
            min_fade_out_time=0.9,
            max_fade_out_time=0.99,
            p=1.0,
        )
        samples_out = augmenter(samples=samples, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples.shape

        rms_after = calculate_rms(samples_out)
        assert rms_after > rms_before

    def test_serialize_parameters(self):
        transform = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "background_noises"), p=1.0
        )
        samples = np.random.normal(0, 1, size=1024).astype(np.float32)
        transform.randomize_parameters(samples, sample_rate=44100)
        json.dumps(transform.serialize_parameters())

    def test_frozen_parameters(self):
        sample_rate = 44100
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
            np.float32
        )
        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_time_between_sounds=2.0,
            max_time_between_sounds=8.0,
            p=1.0,
        )
        samples_out1 = augmenter(samples=samples, sample_rate=sample_rate)

        augmenter.freeze_parameters()
        samples_out2 = augmenter(samples, sample_rate)

        assert_array_equal(samples_out1, samples_out2)

    def test_multichannel_audio_not_supported_yet(self):
        sample_rate = 44100
        samples_chn0 = np.sin(np.linspace(0, 440 * 2 * np.pi, 2 * sample_rate)).astype(
            np.float32
        )
        samples_chn1 = np.sin(np.linspace(0, 440 * 2 * np.pi, 2 * sample_rate)).astype(
            np.float32
        )
        samples = np.vstack((samples_chn0, samples_chn1))

        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_time_between_sounds=2.0,
            max_time_between_sounds=8.0,
            p=1.0,
        )
        with pytest.raises(MultichannelAudioNotSupportedException):
            augmenter(samples=samples, sample_rate=sample_rate)

    def test_picklability(self):
        transform = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"), p=1.0
        )
        pickled = pickle.dumps(transform)
        unpickled = pickle.loads(pickled)
        assert transform.sound_file_paths == unpickled.sound_file_paths

    def test_noise_rms_parameter(self):
        np.random.seed(80085)
        random.seed("ling")
        sample_rate = 44100
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
            np.float32
        )
        rms_before = calculate_rms(samples)
        augmenter_relative_to_whole_input = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_time_between_sounds=2.0,
            max_time_between_sounds=8.0,
            min_snr_db=5,
            max_snr_db=5,
            noise_rms="relative_to_whole_input",
            p=1.0,
        )

        augmenter_absolute = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_time_between_sounds=2.0,
            max_time_between_sounds=8.0,
            noise_rms="absolute",
            p=1.0,
        )

        samples_out_absolute = augmenter_absolute(
            samples=samples, sample_rate=sample_rate
        )
        samples_out_relative_to_whole_input = augmenter_relative_to_whole_input(
            samples=samples, sample_rate=sample_rate
        )
        assert samples_out_absolute.dtype == np.float32
        assert samples_out_absolute.shape == samples.shape
        rms_after_relative_to_whole_path = calculate_rms(
            samples_out_relative_to_whole_input
        )
        rms_after_absolute = calculate_rms(samples_out_absolute)
        assert rms_after_absolute > rms_before
        assert rms_after_relative_to_whole_path > rms_before

    def test_include_silence_in_noise_rms_calculation(self):
        np.random.seed(420)
        random.seed(420)
        sample_rate = 44100
        samples = np.sin(np.linspace(0, 440 * 2 * np.pi, 9 * sample_rate)).astype(
            np.float32
        )
        rms_before = calculate_rms(samples)
        augmenter = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_time_between_sounds=2.0,
            max_time_between_sounds=4.0,
            noise_rms="absolute",
            include_silence_in_noise_rms_estimation=False,
            p=1.0,
        )

        samples_out = augmenter(samples=samples, sample_rate=sample_rate)

        rms_after = calculate_rms(samples_out)
        assert rms_after > rms_before

    def test_add_noises_with_same_level(self):
        dummy_audio = np.random.randint(1, 5, 250000)
        transform_same_noise_level = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_snr_db=15,
            max_snr_db=30,
            noise_rms="relative",
            add_all_noises_with_same_level=True,
            min_time_between_sounds=0.5,
            max_time_between_sounds=1,
            p=1,
        )

        transform_different_noise_level = AddShortNoises(
            sounds_path=os.path.join(DEMO_DIR, "short_noises"),
            min_snr_db=15,
            max_snr_db=30,
            add_all_noises_with_same_level=False,
            min_time_between_sounds=0.5,
            max_time_between_sounds=1,
            p=1,
        )

        for i in range(3):
            transform_same_noise_level.randomize_parameters(
                dummy_audio, sample_rate=44100
            )
            sounds = transform_same_noise_level.parameters["sounds"]
            snr_sounds_same_level = [sounds[j]["snr_db"] for j in range(len(sounds))]
            transform_different_noise_level.randomize_parameters(
                dummy_audio, sample_rate=44100
            )
            sounds = transform_different_noise_level.parameters["sounds"]
            snr_sounds_different_level = [
                sounds[j]["snr_db"] for j in range(len(sounds))
            ]

            assert len(set(snr_sounds_same_level)) == 1
            assert len(set(snr_sounds_different_level)) > 1

    def test_validation(self):
        with pytest.raises(ValueError):
            AddShortNoises(
                sounds_path=os.path.join(DEMO_DIR, "short_noises"),
                min_snr_db=40.0,
                max_snr_db=20.0,
            )



import random
from typing import Any

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseSpectrogramTransform
from audiomentations.core.serialization import get_shortest_class_fullname
from audiomentations.core.utils import format_args

REPR_INDENT_STEP = 2


class BaseCompose:
    def __init__(self, transforms, p: float = 1.0, shuffle: bool = False):
        self.transforms = transforms
        self.p = p
        self.shuffle = shuffle
        self.are_parameters_frozen = False

        name_list = []
        for transform in self.transforms:
            name_list.append(type(transform).__name__)
        self.__name__ = "_".join(name_list)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.indented_repr()

    def randomize_parameters(self, *args, **kwargs):
        """
        Randomize and define parameters of every transform in composition.
        """
        apply_to_children = kwargs.get("apply_to_children", True)
        if apply_to_children:
            if "apply_to_children" in kwargs:
                del kwargs["apply_to_children"]
            for transform in self.transforms:
                transform.randomize_parameters(*args, **kwargs)

    def freeze_parameters(self, apply_to_children=True):
        """
        Mark all parameters as frozen, i.e. do not randomize them for each call. This can be
        useful if you want to apply an effect chain with the exact same parameters to multiple
        sounds.
        """
        self.are_parameters_frozen = True
        if apply_to_children:
            for transform in self.transforms:
                transform.freeze_parameters()

    def unfreeze_parameters(self, apply_to_children=True):
        """
        Unmark all parameters as frozen, i.e. let them be randomized for each call.
        """
        self.are_parameters_frozen = False
        if apply_to_children:
            for transform in self.transforms:
                transform.unfreeze_parameters()

    def indented_repr(self, indent: int = REPR_INDENT_STEP) -> str:
        args = {
            k: v
            for k, v in self.to_dict_private().items()
            if not (k.startswith("__") or k == "transforms")
        }
        repr_string = self.__class__.__name__ + "(["
        for t in self.transforms:
            repr_string += "\n"
            t_repr = (
                t.indented_repr(indent + REPR_INDENT_STEP)
                if hasattr(t, "indented_repr")
                else repr(t)
            )
            repr_string += " " * indent + t_repr + ","
        repr_string += (
            "\n" + " " * (indent - REPR_INDENT_STEP) + f"], {format_args(args)})"
        )
        return repr_string

    @classmethod
    def get_class_fullname(cls) -> str:
        return get_shortest_class_fullname(cls)

    def to_dict_private(self) -> dict[str, Any]:
        return {
            "__class_fullname__": self.get_class_fullname(),
            "p": self.p,
            "transforms": [t.to_dict_private() for t in self.transforms],
        }


class Compose(BaseCompose):
    """
    Compose applies the given sequence of transforms when called,
    optionally shuffling the sequence for every call.

    Usage example:

    ```
    augment = Compose([
        AddGaussianNoise(min_amplitude=0.001, max_amplitude=0.015, p=0.5),
        TimeStretch(min_rate=0.8, max_rate=1.25, p=0.5),
        PitchShift(min_semitones=-4, max_semitones=4, p=0.5),
        Shift(min_shift=-0.5, max_shift=0.5, p=0.5),
    ])

    # Generate 2 seconds of dummy audio for the sake of example
    samples = np.random.uniform(low=-0.2, high=0.2, size=(32000,)).astype(np.float32)

    # Augment/transform/perturb the audio data
    augmented_samples = augment(samples=samples, sample_rate=16000)
    ```
    """

    def __init__(self, transforms, p=1.0, shuffle=False):
        super().__init__(transforms, p, shuffle)

    def __call__(self, samples: NDArray[np.float32], sample_rate: int):
        transforms = self.transforms.copy()
        should_apply = random.random() < self.p
        # TODO: Adhere to self.are_parameters_frozen
        # https://github.com/iver56/audiomentations/issues/135
        if should_apply:
            if self.shuffle:
                random.shuffle(transforms)
            for transform in transforms:
                samples = transform(samples, sample_rate)

        return samples


class SpecCompose(BaseCompose):
    def __init__(self, transforms, p=1.0, shuffle=False):
        super().__init__(transforms, p, shuffle)

    def __call__(self, magnitude_spectrogram):
        transforms = self.transforms.copy()
        should_apply = random.random() < self.p
        # TODO: Adhere to self.are_parameters_frozen
        # https://github.com/iver56/audiomentations/issues/135
        if should_apply:
            if self.shuffle:
                random.shuffle(transforms)
            for transform in transforms:
                magnitude_spectrogram = transform(magnitude_spectrogram)

        return magnitude_spectrogram


class SomeOf(BaseCompose):
    """
    SomeOf randomly picks several of the given transforms when called, and applies these
    transforms. The number of transforms to apply can be chosen in two different ways:

        - Pick exactly n transforms:
            Example:    # pick exactly two of the transforms
                        SomeOf(2, [transform1, transform2, transform3])

        - Pick between a minimum and maximum number of transforms.
            Examples:   # pick 1 to 3 of the transforms
                        SomeOf((1, 3), [transform1, transform2, transform3])

                        # pick 2 to all of the transforms
                        SomeOf((2, None), [transform1, transform2, transform3, transform4])

    Usage example:
    ```
    augment = SomeOf(
        (2, None),
        [
            TimeStretch(min_rate=0.8, max_rate=1.25, p=1.0),
            PitchShift(min_semitones=-4, max_semitones=4, p=1.0),
            Gain(min_gain_in_db=-12, max_gain_in_db=-6, p=1.0),
        ],
    )

    # Generate 2 seconds of dummy audio for the sake of example
    samples = np.random.uniform(low=-0.2, high=0.2, size=(32000,)).astype(np.float32)

    # Augment/transform/perturb the audio data
    augmented_samples = augment(samples=samples, sample_rate=16000)

    # Result: 2 or more transforms will be applied from the list of transforms.
    ```
    """

    def __init__(self, num_transforms: int or tuple, transforms, p: float = 1.0):
        super().__init__(transforms, p)
        self.transform_indexes = []
        self.num_transforms = num_transforms
        self.should_apply = True

    def randomize_parameters(self, *args, **kwargs):
        super().randomize_parameters(*args, **kwargs)
        self.should_apply = random.random() < self.p
        if self.should_apply:
            if type(self.num_transforms) == tuple:
                if self.num_transforms[1] is None:
                    num_transforms_to_apply = random.randint(
                        self.num_transforms[0], len(self.transforms)
                    )
                else:
                    num_transforms_to_apply = random.randint(
                        self.num_transforms[0], self.num_transforms[1]
                    )
            else:
                num_transforms_to_apply = self.num_transforms
            all_transforms_indexes = list(range(len(self.transforms)))
            self.transform_indexes = sorted(
                random.sample(all_transforms_indexes, num_transforms_to_apply)
            )
        return self.transform_indexes

    def __call__(self, *args, **kwargs):
        if not self.are_parameters_frozen:
            kwargs["apply_to_children"] = False
            self.randomize_parameters(*args, **kwargs)

        if self.should_apply:
            if "apply_to_children" in kwargs:
                del kwargs["apply_to_children"]

            if issubclass(type(self.transforms[0]), BaseSpectrogramTransform):
                if "magnitude_spectrogram" in kwargs:
                    magnitude_spectrogram = kwargs["magnitude_spectrogram"]
                else:
                    magnitude_spectrogram = args[0]

                for transform_index in self.transform_indexes:
                    magnitude_spectrogram = self.transforms[transform_index](
                        magnitude_spectrogram
                    )

                return magnitude_spectrogram
            else:  # The transforms are subclasses of BaseWaveformTransform
                if "sample_rate" in kwargs:
                    samples = kwargs["samples"] if "samples" in kwargs else args[0]
                    sample_rate = kwargs["sample_rate"]
                else:
                    samples = args[0]
                    sample_rate = args[1]

                for transform_index in self.transform_indexes:
                    samples = self.transforms[transform_index](samples, sample_rate)

                return samples

        if "samples" in kwargs:
            return kwargs["samples"]
        elif "magnitude_spectrogram" in kwargs:
            return kwargs["magnitude_spectrogram"]
        else:
            return args[0]


class OneOf(BaseCompose):
    """
    OneOf randomly picks one of the given transforms when called, and applies that
    transform.
    Usage example:
    ```
    augment = OneOf([
        TimeStretch(min_rate=0.8, max_rate=1.25, p=1.0),
        PitchShift(min_semitones=-4, max_semitones=4, p=1.0),
    ])
    # Generate 2 seconds of dummy audio for the sake of example
    samples = np.random.uniform(low=-0.2, high=0.2, size=(32000,)).astype(np.float32)
    # Augment/transform/perturb the audio data
    augmented_samples = augment(samples=samples, sample_rate=16000)
    # Result: The audio was either time-stretched or pitch-shifted, but not both
    ```
    """

    def __init__(self, transforms, p: float = 1.0):
        super().__init__(transforms, p)
        self.transform_index = 0
        self.should_apply = True

    def randomize_parameters(self, *args, **kwargs):
        super().randomize_parameters(*args, **kwargs)
        self.should_apply = random.random() < self.p
        if self.should_apply:
            self.transform_index = random.randint(0, len(self.transforms) - 1)

    def __call__(self, *args, **kwargs):
        if not self.are_parameters_frozen:
            kwargs["apply_to_children"] = False
            self.randomize_parameters(*args, **kwargs)

        if self.should_apply:
            if "apply_to_children" in kwargs:
                del kwargs["apply_to_children"]
            return self.transforms[self.transform_index](*args, **kwargs)

        if "samples" in kwargs:
            return kwargs["samples"]
        elif "magnitude_spectrogram" in kwargs:
            return kwargs["magnitude_spectrogram"]
        else:
            return args[0]



import random

import librosa
import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class TimeStretch(BaseWaveformTransform):
    """Time stretch the signal without changing the pitch"""

    supports_multichannel = True

    def __init__(
        self,
        min_rate: float = 0.8,
        max_rate: float = 1.25,
        leave_length_unchanged: bool = True,
        p: float = 0.5,
    ):
        super().__init__(p)
        assert min_rate >= 0.1
        assert max_rate <= 10
        assert min_rate <= max_rate
        self.min_rate = min_rate
        self.max_rate = max_rate
        self.leave_length_unchanged = leave_length_unchanged

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            """
            If rate > 1, then the signal is sped up.
            If rate < 1, then the signal is slowed down.
            """
            self.parameters["rate"] = random.uniform(self.min_rate, self.max_rate)

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        try:
            time_stretched_samples = librosa.effects.time_stretch(
                samples, rate=self.parameters["rate"]
            )
        except librosa.util.exceptions.ParameterError:
            # In librosa<0.9.0 time_stretch doesn't natively support multichannel audio.
            # Here we use a workaround that simply loops over the channels instead.
            # TODO: Remove this workaround when we remove support for librosa<0.9.0
            time_stretched_channels = []
            for i in range(samples.shape[0]):
                time_stretched_samples = librosa.effects.time_stretch(
                    samples[i], rate=self.parameters["rate"]
                )
                time_stretched_channels.append(time_stretched_samples)
            time_stretched_samples = np.array(
                time_stretched_channels, dtype=samples.dtype
            )

        if self.leave_length_unchanged:
            # Apply zero padding if the time stretched audio is not long enough to fill the
            # whole space, or crop the time stretched audio if it ended up too long.
            padded_samples = np.zeros(shape=samples.shape, dtype=samples.dtype)
            window = time_stretched_samples[..., : samples.shape[-1]]
            actual_window_length = window.shape[
                -1
            ]  # may be smaller than samples.shape[-1]
            padded_samples[..., :actual_window_length] = window
            time_stretched_samples = padded_samples
        return time_stretched_samples



import numpy as np
import pytest

from audiomentations import ClippingDistortion, Compose


class TestClippingDistortion:
    def test_distort(self):
        sample_len = 1024
        samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
        sample_rate = 16000
        augmenter = Compose(
            [
                ClippingDistortion(
                    min_percentile_threshold=20, max_percentile_threshold=40, p=1.0
                )
            ]
        )

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert len(samples_out) == sample_len
        assert sum(abs(samples_out)) < sum(abs(samples_in))

    def test_distort_multichannel(self):
        sample_len = 32000
        samples_in = np.random.normal(0, 1, size=(2, sample_len)).astype(np.float32)
        sample_rate = 16000
        augmenter = ClippingDistortion(
            min_percentile_threshold=20, max_percentile_threshold=40, p=1.0
        )

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples_in.shape
        assert np.sum(np.abs(samples_out)) < np.sum(np.abs(samples_in))
        assert np.amax(samples_out[0, :]) == pytest.approx(np.amax(samples_out[1, :]))



import numpy as np
from numpy.testing import assert_array_equal

from audiomentations import SpecChannelShuffle, SpecFrequencyMask, SpecCompose


class TestSpecCompose:
    def test_freeze_and_unfreeze_parameters(self):
        spectrogram = np.random.random((256, 256, 3))
        augmenter = SpecCompose(
            [
                SpecChannelShuffle(p=1.0),
                SpecFrequencyMask(p=1.0),
            ]
        )
        perturbed_samples1 = augmenter(magnitude_spectrogram=spectrogram)
        augmenter.freeze_parameters()
        for transform in augmenter.transforms:
            assert transform.are_parameters_frozen == True
        perturbed_samples2 = augmenter(magnitude_spectrogram=spectrogram)

        assert_array_equal(perturbed_samples1, perturbed_samples2)

        augmenter.unfreeze_parameters()
        for transform in augmenter.transforms:
            assert transform.are_parameters_frozen == False

    def test_randomize_parameters_and_apply(self):
        spectrogram = np.random.random((256, 256, 3))
        augmenter = SpecCompose(
            [
                SpecChannelShuffle(p=1.0),
                SpecFrequencyMask(p=1.0),
            ]
        )
        augmenter.freeze_parameters()
        augmenter.randomize_parameters(magnitude_spectrogram=spectrogram)

        parameters = [transform.parameters for transform in augmenter.transforms]

        perturbed_samples1 = augmenter(magnitude_spectrogram=spectrogram)
        perturbed_samples2 = augmenter(magnitude_spectrogram=spectrogram)

        assert_array_equal(perturbed_samples1, perturbed_samples2)

        augmenter.unfreeze_parameters()

        for transform_parameters, transform in zip(parameters, augmenter.transforms):
            assert transform_parameters == transform.parameters
            assert transform.are_parameters_frozen == False



import random
import warnings

import librosa
import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class PitchShift(BaseWaveformTransform):
    """Pitch shift the sound up or down without changing the tempo"""

    supports_multichannel = True

    def __init__(
        self, min_semitones: float = -4.0, max_semitones: float = 4.0, p: float = 0.5
    ):
        """
        :param min_semitones: Minimum semitones to shift. A negative number means shift down.
        :param max_semitones: Maximum semitones to shift. A positive number means shift up.
        :param p: The probability of applying this transform
        """
        super().__init__(p)
        assert min_semitones >= -12
        assert max_semitones <= 12
        assert min_semitones <= max_semitones
        self.min_semitones = min_semitones
        self.max_semitones = max_semitones

    def randomize_parameters(self, samples: NDArray[np.float32], sample_rate: int):
        super().randomize_parameters(samples, sample_rate)
        if self.parameters["should_apply"]:
            self.parameters["num_semitones"] = random.uniform(
                self.min_semitones, self.max_semitones
            )

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        try:
            resample_type = (
                "kaiser_best" if librosa.__version__.startswith("0.8.") else "soxr_hq"
            )
            pitch_shifted_samples = librosa.effects.pitch_shift(
                samples,
                sr=sample_rate,
                n_steps=self.parameters["num_semitones"],
                res_type=resample_type,
            )
        except librosa.util.exceptions.ParameterError:
            warnings.warn(
                "Warning: You are probably using an old version of librosa. Upgrade"
                " librosa to 0.9.0 or later for better performance when applying"
                " PitchShift to stereo audio."
            )
            # In librosa<0.9.0 pitch_shift doesn't natively support multichannel audio.
            # Here we use a workaround that simply loops over the channels instead.
            # TODO: Remove this workaround when we remove support for librosa<0.9.0
            pitch_shifted_samples = np.copy(samples)
            for i in range(samples.shape[0]):
                pitch_shifted_samples[i] = librosa.effects.pitch_shift(
                    pitch_shifted_samples[i],
                    sr=sample_rate,
                    n_steps=self.parameters["num_semitones"],
                )

        return pitch_shifted_samples



import numpy as np
from numpy.testing import assert_almost_equal

from audiomentations import PolarityInversion, Compose


class TestPolarityInversion:
    def test_polarity_inversion(self):
        samples = np.array([1.0, 0.5, -0.25, -0.125, 0.0], dtype=np.float32)
        sample_rate = 16000

        augment = Compose([PolarityInversion(p=1.0)])
        inverted_samples = augment(samples=samples, sample_rate=sample_rate)
        assert_almost_equal(
            inverted_samples, np.array([-1.0, -0.5, 0.25, 0.125, 0.0], dtype=np.float32)
        )
        assert inverted_samples.dtype == np.float32

    def test_polarity_inversion_multichannel(self):
        samples = np.array(
            [[1.0, 0.5, -0.25, -0.125, 0.0], [1.0, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        sample_rate = 16000

        augment = Compose([PolarityInversion(p=1.0)])
        inverted_samples = augment(samples=samples, sample_rate=sample_rate)
        assert_almost_equal(
            inverted_samples,
            np.array(
                [[-1.0, -0.5, 0.25, 0.125, 0.0], [-1.0, -0.5, 0.25, 0.125, 0.0]],
                dtype=np.float32,
            ),
        )
        assert inverted_samples.dtype == np.float32



import pytest
import numpy as np
from numpy.testing import assert_almost_equal

from audiomentations import LoudnessNormalization


class TestLoudnessNormalization:
    def test_loudness_normalization(self):
        samples = np.random.uniform(low=-0.2, high=-0.001, size=(8000,)).astype(
            np.float32
        )
        sample_rate = 16000

        augment = LoudnessNormalization(min_lufs=-32, max_lufs=-12, p=1.0)
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        gain_factors = processed_samples / samples
        assert np.amin(gain_factors) == pytest.approx(np.amax(gain_factors))
        assert processed_samples.dtype == np.float32

    def test_loudness_normalization_digital_silence(self):
        samples = np.zeros(8000, dtype=np.float32)
        sample_rate = 16000

        augment = LoudnessNormalization(min_lufs=-32, max_lufs=-12, p=1.0)
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert_almost_equal(processed_samples, np.zeros(8000, dtype=np.float32))
        assert processed_samples.dtype == np.float32

    def test_loudness_normalization_too_short_input(self):
        samples = np.random.uniform(low=-0.2, high=-0.001, size=(800,)).astype(
            np.float32
        )
        sample_rate = 16000

        augment = LoudnessNormalization(min_lufs=-32, max_lufs=-12, p=1.0)
        with pytest.raises(ValueError):
            _ = augment(samples=samples, sample_rate=sample_rate)

    def test_loudness_normalization_multichannel(self):
        samples = np.random.uniform(low=-0.2, high=-0.001, size=(3, 8000)).astype(
            np.float32
        )
        sample_rate = 16000

        augment = LoudnessNormalization(min_lufs=-32, max_lufs=-12, p=1.0)
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        gain_factors = processed_samples / samples
        assert np.amin(gain_factors) == pytest.approx(np.amax(gain_factors))
        assert processed_samples.dtype == np.float32

    def test_validation(self):
        with pytest.raises(ValueError):
            LoudnessNormalization(
                min_lufs=-12.0,
                max_lufs=-32.0,
            )



import math
import numpy as np
import pytest

from audiomentations import Resample, Compose


class TestResample:
    @pytest.mark.parametrize(
        "samples",
        [
            np.zeros((512,), dtype=np.float32),
            np.zeros(
                (
                    2,
                    2512,
                ),
                dtype=np.float32,
            ),
        ],
    )
    def test_resample(self, samples):
        sample_rate = 16000
        augmenter = Compose(
            [Resample(min_sample_rate=8000, max_sample_rate=44100, p=1.0)]
        )
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert samples.shape[-1] <= math.ceil(samples.shape[-1] * 44100 / sample_rate)
        assert samples.shape[-1] >= math.ceil(samples.shape[-1] * 8000 / sample_rate)



import codecs
import os
import re

from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    long_description = readme_file.read()

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="audiomentations",
    version=find_version("audiomentations", "__init__.py"),
    author="Iver Jordal",
    description=(
        "A Python library for audio data augmentation. Inspired by albumentations."
        " Useful for machine learning."
    ),
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iver56/audiomentations",
    packages=find_packages(exclude=["demo", "tests"]),
    install_requires=[
        "numpy>=1.21.0,<2",
        "numpy-minmax>=0.3.0,<1",
        "numpy-rms>=0.4.2,<1",
        "librosa>=0.8.0,!=0.10.0,<0.11.0",
        "scipy>=1.4,<1.13",
        "soxr>=0.3.2,<1.0.0",
    ],
    extras_require={
        "extras": [
            "cylimiter==0.3.0",
            "lameenc>=1.2.0,<2",
            "pydub>=0.22.0,<1",
            "pyloudnorm>=0.1.0",
            "pyroomacoustics>=0.6.0",
        ]
    },
    python_requires=">=3.8,<3.13",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    project_urls={
        "Homepage": "https://github.com/iver56/audiomentations",
        "Documentation": "https://iver56.github.io/audiomentations/",
        "Changelog": "https://iver56.github.io/audiomentations/changelog/",
        "Issue Tracker": "https://github.com/iver56/audiomentations/issues",
    },
)



import numpy as np
import pytest

from audiomentations import TimeMask, Compose


class TestTimeMask:
    def test_apply_time_mask(self):
        sample_len = 1024
        samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
        sample_rate = 16000
        augmenter = Compose([TimeMask(min_band_part=0.2, max_band_part=0.5, p=1.0)])

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert len(samples_out) == sample_len

        std_in = np.mean(np.abs(samples_in))
        std_out = np.mean(np.abs(samples_out))
        assert std_out < std_in

    def test_invalid_params(self):
        with pytest.raises(ValueError):
            TimeMask(min_band_part=0.5, max_band_part=1.5)

        with pytest.raises(ValueError):
            TimeMask(min_band_part=-0.5, max_band_part=0.5)

        with pytest.raises(ValueError):
            TimeMask(min_band_part=0.6, max_band_part=0.5)

    def test_apply_time_mask_multichannel(self):
        sample_len = 1024
        samples_in = np.random.normal(0, 1, size=(2, sample_len)).astype(np.float32)
        sample_rate = 16000
        augmenter = TimeMask(min_band_part=0.2, max_band_part=0.5, p=1.0)

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert samples_out.shape == samples_in.shape

        std_in = np.mean(np.abs(samples_in))
        std_out = np.mean(np.abs(samples_out))
        assert std_out < std_in

    def test_apply_time_mask_with_fade(self):
        sample_len = 1024
        samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
        sample_rate = 16000
        augmenter = Compose(
            [TimeMask(min_band_part=0.2, max_band_part=0.5, fade=True, p=1.0)]
        )

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert len(samples_out) == sample_len

        std_in = np.mean(np.abs(samples_in))
        std_out = np.mean(np.abs(samples_out))
        assert std_out < std_in

    def test_apply_time_mask_with_fade_short_signal(self):
        sample_len = 100
        samples_in = np.random.normal(0, 1, size=sample_len).astype(np.float32)
        sample_rate = 16000
        augmenter = Compose(
            [TimeMask(min_band_part=0.2, max_band_part=0.5, fade=True, p=1.0)]
        )

        samples_out = augmenter(samples=samples_in, sample_rate=sample_rate)
        assert samples_out.dtype == np.float32
        assert len(samples_out) == sample_len

        std_in = np.mean(np.abs(samples_in))
        std_out = np.mean(np.abs(samples_out))
        assert std_out < std_in



import random

import numpy as np
import pytest
from numpy import array_equal

from audiomentations import Padding


class TestPadding:
    @pytest.mark.parametrize("mode", ["silence", "wrap", "reflect"])
    @pytest.mark.parametrize("pad_section", ["start", "end"])
    def test_padding_mono_1d(self, mode, pad_section):
        random.seed(546)
        samples = np.array([0.5, 0.6, -0.2, 1.0], dtype=np.float32)
        sample_rate = 16000
        input_shape = samples.shape
        augmenter = Padding(mode=mode, pad_section=pad_section, p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert samples.shape == input_shape

    def test_padding_mono_2d(self):
        samples = np.array(
            [[0.9, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        sample_rate = 16000
        input_shape = samples.shape

        augmenter = Padding(p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert samples.shape == input_shape

    @pytest.mark.parametrize("mode", ["silence", "wrap", "reflect"])
    @pytest.mark.parametrize("pad_section", ["start", "end"])
    def test_padding_multichannel(self, mode, pad_section):
        samples = np.array(
            [
                [0.9, 0.5, -0.25, -0.125, 0.0],
                [0.95, 0.5, -0.25, -0.125, 0.0],
                [0.95, 0.5, -0.25, -0.125, 0.0],
            ],
            dtype=np.float32,
        )
        sample_rate = 16000
        input_shape = samples.shape

        augmenter = Padding(mode=mode, pad_section=pad_section, p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert samples.dtype == np.float32
        assert samples.shape == input_shape

    def test_padding_reflect_start(self):
        samples = np.array([0.5, 0.6, 0.9, -0.2, 1.0], dtype=np.float32)
        sample_rate = 16000
        augmenter = Padding(
            mode="reflect",
            pad_section="start",
            min_fraction=0.4,
            max_fraction=0.4,
            p=1.0,
        )
        samples = augmenter(samples=samples, sample_rate=sample_rate)
        assert array_equal(
            samples, np.array([1.0, -0.2, 0.9, -0.2, 1.0], dtype=np.float32)
        )

    def test_padding_reflect_end(self):
        samples = np.array([0.5, 0.6, 0.9, -0.2, 1.0], dtype=np.float32)
        sample_rate = 16000
        augmenter = Padding(
            mode="reflect",
            pad_section="end",
            min_fraction=0.4,
            max_fraction=0.4,
            p=1.0,
        )
        samples = augmenter(samples=samples, sample_rate=sample_rate)
        assert array_equal(
            samples, np.array([0.5, 0.6, 0.9, 0.6, 0.5], dtype=np.float32)
        )

    def test_pad_nothing(self):
        samples = np.array([0.5, 0.6, -0.2, 0.1], dtype=np.float32)
        sample_rate = 16000
        input_shape = samples.shape
        augmenter = Padding(min_fraction=0.0, max_fraction=0.0, p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert array_equal(samples, np.array([0.5, 0.6, -0.2, 0.1], dtype=np.float32))
        assert samples.dtype == np.float32
        assert samples.shape == input_shape

    def test_pad_everything(self):
        samples = np.array([0.5, 0.6, -0.2, 0.7], dtype=np.float32)
        sample_rate = 16000
        input_shape = samples.shape
        augmenter = Padding(min_fraction=1.0, max_fraction=1.0, p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert not np.any(samples)
        assert samples.dtype == np.float32
        assert samples.shape == input_shape



from typing import Callable

import numpy as np
from numpy.typing import NDArray

from audiomentations.core.transforms_interface import BaseWaveformTransform


class Lambda(BaseWaveformTransform):
    """
    Apply a user-defined transform (callable) to the signal.
    """

    supports_multichannel = True

    def __init__(self, transform: Callable, p: float = 0.5, **kwargs):
        """
        :param transform: A callable to be applied over samples. It should input
            samples (ndarray), sample_rate (int) and optionally some user-defined
            keyword arguments.
        :param p: The probability of applying this transform
        :param **kwargs: Any extra keyword arguments to be passed to the transform.
        """
        super().__init__(p=p)
        self.transform = transform
        self.kwargs = kwargs

    def apply(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        return self.transform(samples, sample_rate, **self.kwargs)



import numpy as np
from numpy.testing import assert_array_equal

from audiomentations import Normalize


class TestNormalize:
    def test_normalize_positive_peak(self):
        samples = np.array([0.5, 0.6, -0.2, 0.0], dtype=np.float32)
        sample_rate = 16000
        augmenter = Normalize(p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert np.amax(samples) == 1.0
        assert samples.dtype == np.float32
        assert samples.shape[-1] == 4

    def test_normalize_negative_peak(self):
        samples = np.array([0.5, 0.6, -0.8, 0.0], dtype=np.float32)
        sample_rate = 16000
        augmenter = Normalize(p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert np.amin(samples) == -1.0
        assert samples[-1] == 0.0
        assert samples.dtype == np.float32
        assert samples.shape[-1] == 4

    def test_normalize_all_zeros(self):
        samples = np.array([0.0, 0.0, 0.0, 0.0], dtype=np.float32)
        sample_rate = 16000
        augmenter = Normalize(p=1.0)
        samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert np.amin(samples) == 0.0
        assert samples[-1] == 0.0
        assert samples.dtype == np.float32
        assert samples.shape[-1] == 4

    def test_normalize_multichannel(self):
        samples = np.array(
            [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        sample_rate = 16000
        augmenter = Normalize(p=1.0)
        processed_samples = augmenter(samples=samples, sample_rate=sample_rate)

        assert_array_equal(processed_samples, samples / 0.95)
        assert processed_samples.dtype == np.float32

    def test_normalize_multichannel_conditionally(self):
        sample_rate = 16000
        augmenter = Normalize(apply_to="only_too_loud_sounds", p=1.0)

        samples = np.array(
            [[0.9, 0.5, -0.25, -0.125, 0.0], [0.95, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        processed_samples = augmenter(samples=samples, sample_rate=sample_rate)
        assert_array_equal(processed_samples, samples)
        assert processed_samples.dtype == np.float32

        samples_too_loud = np.array(
            [[0.9, 0.5, -0.25, -0.125, 0.0], [1.2, 0.5, -0.25, -0.125, 0.0]],
            dtype=np.float32,
        )
        processed_samples = augmenter(samples=samples_too_loud, sample_rate=sample_rate)
        assert_array_equal(processed_samples, samples_too_loud / 1.2)
        assert processed_samples.dtype == np.float32



import random

from audiomentations.core.transforms_interface import BaseSpectrogramTransform


class SpecChannelShuffle(BaseSpectrogramTransform):
    """
    Shuffle the channels of a multichannel spectrogram (channels last).
    This can help combat positional bias.
    """
    supports_multichannel = True
    supports_mono = False

    def randomize_parameters(self, magnitude_spectrogram):
        super().randomize_parameters(magnitude_spectrogram)
        if self.parameters["should_apply"]:
            self.parameters["shuffled_channel_indexes"] = list(range(magnitude_spectrogram.shape[-1]))
            random.shuffle(self.parameters["shuffled_channel_indexes"])

    def apply(self, magnitude_spectrogram):
        return magnitude_spectrogram[..., self.parameters["shuffled_channel_indexes"]]



import json
import pytest

import numpy as np

from audiomentations import AddGaussianSNR


class TestGaussianSNR:
    def test_gaussian_noise_snr_defaults(self):
        np.random.seed(42)
        samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
        augmenter = AddGaussianSNR(p=1.0)
        std_in = np.mean(np.abs(samples_in))
        samples_out = augmenter(samples=samples_in, sample_rate=16000)
        std_out = np.mean(np.abs(samples_out))
        assert samples_out.dtype == np.float32
        assert not (float(std_out) == pytest.approx(0.0))
        assert std_out > std_in

    def test_gaussian_noise_snr(self):
        np.random.seed(42)
        samples_in = np.random.normal(0, 1, size=1024).astype(np.float32)
        augmenter = AddGaussianSNR(min_snr_db=15, max_snr_db=35, p=1.0)
        std_in = np.mean(np.abs(samples_in))
        samples_out = augmenter(samples=samples_in, sample_rate=16000)
        std_out = np.mean(np.abs(samples_out))
        assert samples_out.dtype == np.float32
        assert not (float(std_out) == pytest.approx(0.0))
        assert std_out > std_in

    def test_serialize_parameters(self):
        np.random.seed(42)
        transform = AddGaussianSNR(min_snr_db=15, max_snr_db=35, p=1.0)
        samples = np.random.normal(0, 1, size=1024).astype(np.float32)
        transform.randomize_parameters(samples, sample_rate=16000)
        json.dumps(transform.serialize_parameters())

    def test_gaussian_noise_snr_multichannel(self):
        np.random.seed(42)
        samples = np.random.normal(0, 0.1, size=(3, 8888)).astype(np.float32)
        augmenter = AddGaussianSNR(min_snr_db=15, max_snr_db=35, p=1.0)
        samples_out = augmenter(samples=samples, sample_rate=16000)

        assert samples_out.dtype == np.float32
        assert float(np.sum(np.abs(samples_out))) > float(np.sum(np.abs(samples)))

    def test_validation(self):
        with pytest.raises(ValueError):
            AddGaussianSNR(
                min_snr_db=40.0,
                max_snr_db=20.0,
            )



import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from audiomentations import Gain
from audiomentations.core.transforms_interface import WrongMultichannelAudioShape


def test_gain():
    samples = np.array([1.0, 0.5, -0.25, -0.125, 0.0], dtype=np.float32)
    sample_rate = 16000

    augment = Gain(min_gain_db=-6, max_gain_db=-6, p=1.0)
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert_almost_equal(
        processed_samples,
        np.array([0.5011872, 0.2505936, -0.1252968, -0.0626484, 0.0], dtype=np.float32),
    )
    assert processed_samples.dtype == np.float32


def test_gain_multichannel():
    samples = np.array(
        [[1.0, 0.5, -0.25, -0.125, 0.0], [1.0, 0.5, -0.25, -0.125, 0.0]],
        dtype=np.float32,
    )
    sample_rate = 16000

    augment = Gain(min_gain_db=-6, max_gain_db=-6, p=1.0)
    processed_samples = augment(samples=samples, sample_rate=sample_rate)
    assert_almost_equal(
        processed_samples,
        np.array(
            [
                [0.5011872, 0.2505936, -0.1252968, -0.0626484, 0.0],
                [0.5011872, 0.2505936, -0.1252968, -0.0626484, 0.0],
            ],
            dtype=np.float32,
        ),
    )
    assert processed_samples.dtype == np.float32


def test_gain_multichannel_with_wrong_dimension_ordering():
    samples = np.random.uniform(low=-0.5, high=0.5, size=(2000, 2)).astype(np.float32)

    augment = Gain(min_gain_db=-6, max_gain_db=-6, p=1.0)

    with pytest.raises(WrongMultichannelAudioShape):
        augment(samples=samples, sample_rate=16000)


def test_gain_min_greater_than_max():
    with pytest.raises(ValueError):
        Gain(min_gain_db=60, max_gain_db=10, p=1.0)



from typing import Callable

import numpy as np
import sys
from numpy.typing import NDArray

from audiomentations import Normalize
from audiomentations.core.utils import (
    calculate_rms,
    convert_decibels_to_amplitude_ratio,
)


class PostGain:
    """
    Gain up or down the audio after the given transform (or set of transforms) has
    processed the audio. There are several methods that determine how the audio should
    be gained. PostGain can be useful for compensating for any gain differences introduced
    by a (set of) transform(s), or for preventing clipping in the output.
    """

    def __init__(
        self,
        transform: Callable[[NDArray[np.float32], int], NDArray[np.float32]],
        method: str,  # , **kwargs
    ):
        """
        :param transform: A callable to be applied. It should input
            samples (ndarray), sample_rate (int) and optionally some user-defined
            keyword arguments.
        :param method: "same_rms", "same_lufs", "peak_normalize_always" or "peak_normalize_always"
        """
        self.transform = transform
        self.method = method
        assert self.method in (
            "same_rms",
            "same_lufs",
            "peak_normalize_always",
            "peak_normalize_if_too_loud",
            # "target_rms",
            # "target_lufs",
            # "target_peak_dbfs",
            # "target_true_peak_dbfs",
        )
        # if self.method == "target_rms":
        #     self.target_rms = kwargs["target_rms"]
        # elif self.method == "target_lufs":
        #     self.target_lufs = kwargs["target_lufs"]
        # elif self.method == "target_peak_dbfs":
        #     self.target_peak_dbfs = kwargs["target_peak_dbfs"]

    def method_same_rms(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        rms_before = calculate_rms(samples)
        samples = self.transform(samples, sample_rate)
        rms_after = calculate_rms(samples)
        gain_factor = rms_before / rms_after
        samples *= gain_factor
        return samples

    def method_same_lufs(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        try:
            import pyloudnorm
        except ImportError:
            print(
                (
                    "Failed to import pyloudnorm. Maybe it is not installed? "
                    "To install the optional pyloudnorm dependency of audiomentations,"
                    " do `pip install audiomentations[extras]` or simply "
                    " `pip install pyloudnorm`"
                ),
                file=sys.stderr,
            )
            raise

        meter = pyloudnorm.Meter(sample_rate)  # create BS.1770 meter
        # transpose because pyloudnorm expects shape like (smp, chn), not (chn, smp)
        lufs_before = meter.integrated_loudness(samples.transpose())
        samples = self.transform(samples, sample_rate)
        lufs_after = meter.integrated_loudness(samples.transpose())
        gain_db = lufs_before - lufs_after
        samples *= convert_decibels_to_amplitude_ratio(gain_db)
        return samples

    def method_peak_normalize_always(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        samples = self.transform(samples, sample_rate)
        return Normalize(apply_to="all", p=1.0)(samples, sample_rate)

    def method_peak_normalize_if_too_loud(
        self, samples: NDArray[np.float32], sample_rate: int
    ) -> NDArray[np.float32]:
        samples = self.transform(samples, sample_rate)
        return Normalize(apply_to="only_too_loud_sounds", p=1.0)(samples, sample_rate)

    def __call__(self, samples: NDArray[np.float32], sample_rate: int) -> NDArray[np.float32]:
        if self.method == "same_rms":
            return self.method_same_rms(samples, sample_rate)
        elif self.method == "same_lufs":
            return self.method_same_lufs(samples, sample_rate)
        elif self.method == "peak_normalize_always":
            return self.method_peak_normalize_always(samples, sample_rate)
        elif self.method == "peak_normalize_if_too_loud":
            return self.method_peak_normalize_if_too_loud(samples, sample_rate)
        else:
            raise Exception()



from copy import deepcopy

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from audiomentations import Gain, AdjustDuration, RepeatPart, Shift
from audiomentations.core.utils import get_crossfade_mask_pair
from tests.test_filter_transforms import get_chirp_test


def adapt_ndim(samples, ndim):
    if samples.ndim < ndim:
        samples = samples[np.newaxis, :]
    return samples


def assert_high_frequency_energy_absence(
    audio_array,
    sample_rate,
    frequency_threshold: float = 10_000.0,
    energy_threshold: float = 0.15,
):
    spectrum = np.fft.rfft(audio_array)
    freqs = np.fft.rfftfreq(len(audio_array), 1 / sample_rate)
    assert np.all(
        np.abs(spectrum[freqs > frequency_threshold]) < energy_threshold
    ), "Detected energy in frequencies above the given frequency_threshold!"


class TestRepeatPart:
    def test_replace_one_repeat(self):
        augment = RepeatPart(mode="replace", crossfade_duration=0.0, p=1.0)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 1,
            "part_start_index": 1,
        }
        augment.freeze_parameters()

        for ndim in (1, 2):
            samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                dtype=np.float32,
            )
            samples = adapt_ndim(samples, ndim)
            processed_samples = augment(samples=samples, sample_rate=4000)
            target_samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.7, 0.8, 0.9, 1.0],
                dtype=np.float32,
            )
            target_samples = adapt_ndim(target_samples, ndim)
            assert_array_almost_equal(processed_samples, target_samples)
            assert processed_samples.shape == samples.shape
            assert processed_samples.dtype == np.float32

    def test_replace_one_repeat_transformed(self):
        augment = RepeatPart(
            mode="replace",
            crossfade_duration=0.0,
            part_transform=Gain(min_gain_db=-6.0, max_gain_db=-6.0, p=1.0),
            p=1.0,
        )
        part_gain_factor = 0.5011872336272722
        augment.part_transform.parameters = {
            "should_apply": True,
            "amplitude_ratio": part_gain_factor,  # -6 dB
        }
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 1,
            "part_start_index": 1,
        }
        augment.freeze_parameters()

        for ndim in (1, 2):
            samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                dtype=np.float32,
            )
            samples = adapt_ndim(samples, ndim)

            processed_samples = augment(samples=samples, sample_rate=4000)

            target_samples = np.array(
                [
                    0.0,
                    0.1,
                    0.2,
                    0.3,
                    0.1 * part_gain_factor,
                    0.2 * part_gain_factor,
                    0.3 * part_gain_factor,
                    0.7,
                    0.8,
                    0.9,
                    1.0,
                ],
                dtype=np.float32,
            )
            target_samples = adapt_ndim(target_samples, ndim)
            assert_array_almost_equal(processed_samples, target_samples)
            assert processed_samples.shape == samples.shape
            assert processed_samples.dtype == np.float32

    def test_freeze_and_unfreeze_part_transform_parameters(self):
        augment = RepeatPart(
            min_part_duration=0.1,
            max_part_duration=0.2,
            part_transform=Gain(p=1.0),
            crossfade_duration=0.0,
            p=1.0,
        )
        dummy_samples = np.zeros(40, dtype=np.float32)
        augment(dummy_samples, 40)
        params1 = deepcopy(augment.part_transform.parameters)
        augment.freeze_parameters()
        augment(dummy_samples, 40)
        params2 = deepcopy(augment.part_transform.parameters)
        assert params1 == params2

        augment.unfreeze_parameters()
        augment(dummy_samples, 40)
        params3 = deepcopy(augment.part_transform.parameters)
        assert params3 != params2

    def test_replace_one_repeat_near_end(self):
        augment = RepeatPart(mode="replace", crossfade_duration=0.0, p=1.0)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 1,
            "part_start_index": 7,
        }
        augment.freeze_parameters()

        for ndim in (1, 2):
            samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                dtype=np.float32,
            )
            samples = adapt_ndim(samples, ndim)
            processed_samples = augment(samples=samples, sample_rate=4000)
            target_samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.7],
                dtype=np.float32,
            )
            target_samples = adapt_ndim(target_samples, ndim)
            assert_array_almost_equal(processed_samples, target_samples)
            assert processed_samples.shape == samples.shape
            assert processed_samples.dtype == np.float32

    def test_replace_two_repeats(self):
        augment = RepeatPart(mode="replace", crossfade_duration=0.0, p=1.0)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 2,
            "part_start_index": 1,
        }
        augment.freeze_parameters()

        for ndim in (1, 2):
            samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                dtype=np.float32,
            )
            samples = adapt_ndim(samples, ndim)
            processed_samples = augment(samples=samples, sample_rate=4000)
            target_samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 1.0],
                dtype=np.float32,
            )
            target_samples = adapt_ndim(target_samples, ndim)
            assert_array_almost_equal(processed_samples, target_samples)
            assert processed_samples.shape == samples.shape
            assert processed_samples.dtype == np.float32

    def test_replace_many_repeats_exceed_input_length(self):
        augment = RepeatPart(mode="replace", crossfade_duration=0.0, p=1.0)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 9,
            "part_start_index": 1,
        }
        augment.freeze_parameters()

        for ndim in (1, 2):
            samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                dtype=np.float32,
            )
            samples = adapt_ndim(samples, ndim)
            processed_samples = augment(samples=samples, sample_rate=4000)
            target_samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1],
                dtype=np.float32,
            )
            target_samples = adapt_ndim(target_samples, ndim)
            assert_array_almost_equal(processed_samples, target_samples)
            assert processed_samples.shape == samples.shape
            assert processed_samples.dtype == np.float32

    def test_replace_many_transformed_repeats_exceed_input_length(self):
        augment = RepeatPart(
            mode="replace",
            crossfade_duration=0.0,
            part_transform=Gain(min_gain_db=-6.0, max_gain_db=-6.0, p=1.0),
            p=1.0,
        )
        part_gain_factor = 0.5011872336272722
        augment.part_transform.parameters = {
            "should_apply": True,
            "amplitude_ratio": part_gain_factor,  # -6 dB
        }
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 9,
            "part_start_index": 1,
        }
        augment.freeze_parameters()

        for ndim in (1, 2):
            samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                dtype=np.float32,
            )
            samples = adapt_ndim(samples, ndim)
            processed_samples = augment(samples=samples, sample_rate=4000)
            target_samples = np.array(
                [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1],
                dtype=np.float32,
            )
            target_samples[4:] *= part_gain_factor
            target_samples = adapt_ndim(target_samples, ndim)
            assert_array_almost_equal(processed_samples, target_samples)
            assert processed_samples.shape == samples.shape
            assert processed_samples.dtype == np.float32

    def test_insert_one_repeat(self):
        augment = RepeatPart(mode="insert", crossfade_duration=0.0, p=1.0)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 1,
            "part_start_index": 1,
        }
        augment.freeze_parameters()

        samples = np.array(
            [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0], dtype=np.float32
        )
        processed_samples = augment(samples=samples, sample_rate=4000)
        assert_array_almost_equal(
            processed_samples,
            np.array(
                [0.0, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                dtype=np.float32,
            ),
        )
        assert processed_samples.dtype == np.float32

    def test_insert_two_repeats(self):
        augment = RepeatPart(mode="insert", crossfade_duration=0.0, p=1.0)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 2,
            "part_start_index": 0,
        }
        augment.freeze_parameters()

        samples = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7], dtype=np.float32)
        processed_samples = augment(samples=samples, sample_rate=4000)
        assert_array_almost_equal(
            processed_samples,
            np.array(
                [0.0, 0.1, 0.2, 0.0, 0.1, 0.2, 0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7],
                dtype=np.float32,
            ),
        )
        assert processed_samples.dtype == np.float32

    def test_insert_two_elongated_repeats(self):
        augment = RepeatPart(
            mode="insert",
            crossfade_duration=0.0,
            part_transform=AdjustDuration(duration_samples=6, p=1.0),
            p=1.0,
        )
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": 3,
            "repeats": 2,
            "part_start_index": 0,
        }
        augment.freeze_parameters()

        samples = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7], dtype=np.float32)
        processed_samples = augment(samples=samples, sample_rate=4000)
        assert_array_almost_equal(
            processed_samples,
            np.array(
                [
                    0.0,
                    0.1,
                    0.2,
                    0.0,  # first repetition starts here
                    0.1,
                    0.2,
                    0.0,  # start of zero padding
                    0.0,
                    0.0,  # end of zero padding
                    0.0,  # second repetition starts here
                    0.1,
                    0.2,
                    0.0,  # start of zero padding
                    0.0,
                    0.0,  # end of zero padding
                    0.3,
                    0.4,
                    0.5,
                    0.6,
                    0.7,
                ],
                dtype=np.float32,
            ),
        )
        assert processed_samples.dtype == np.float32

    def test_insert_two_repeats_with_crossfading(self):
        augment = RepeatPart(mode="insert", crossfade_duration=0.005, p=1.0)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": int(0.25 * 44100),
            "repeats": 2,
            "part_start_index": 150,
        }
        augment.freeze_parameters()

        sample_rate = 44100
        samples = get_chirp_test(sample_rate, duration=2) * 0.3
        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        assert processed_samples.dtype == np.float32

    def test_insert_two_repeats_with_crossfading_and_part_transform(self):
        # The purpose of this test is to cover two different cases:
        # 1) The fade from the last part to the tail is correlated (use equal-gain fade)
        # 2) The fade from the last part to the tail is uncorrelated (use equal-energy fade)
        for part_transform in [Gain(p=0.0), Shift(p=1.0)]:
            augment = RepeatPart(
                mode="insert",
                crossfade_duration=0.005,
                part_transform=part_transform,
                p=1.0,
            )
            augment.parameters = {
                "should_apply": True,
                "part_num_samples": int(0.25 * 44100),
                "repeats": 2,
                "part_start_index": 150,
            }
            augment.freeze_parameters()

            sample_rate = 44100
            samples = get_chirp_test(sample_rate, duration=2) * 0.3
            processed_samples = augment(samples=samples, sample_rate=sample_rate)

            assert processed_samples.dtype == np.float32

    def test_replace_mode_two_repeats_with_crossfading(self):
        augment = RepeatPart(mode="replace", crossfade_duration=0.005, p=1.0)
        part_num_samples = int(0.25 * 44100)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": part_num_samples,
            "repeats": 2,
            "part_start_index": 142,
        }
        augment.freeze_parameters()

        sample_rate = 44100
        amplitude = 0.3
        samples = get_chirp_test(sample_rate, duration=2) * amplitude
        processed_samples = augment(samples=samples, sample_rate=sample_rate)

        assert processed_samples.dtype == np.float32

        augment_without_crossfade = RepeatPart(
            mode="replace", crossfade_duration=0.000, p=1.0
        )
        augment_without_crossfade.parameters = augment.parameters
        augment_without_crossfade.freeze_parameters()
        processed_samples_without_crossfade = augment_without_crossfade(
            samples=samples, sample_rate=sample_rate
        )

        idx0 = augment.parameters["part_start_index"] + part_num_samples - 1
        idx1 = idx0 + 1
        seam_impulse_magnitude_without_crossfade = abs(
            processed_samples_without_crossfade[..., idx1]
            - processed_samples_without_crossfade[..., idx0]
        )
        assert seam_impulse_magnitude_without_crossfade > amplitude  # audible click!
        seam_impulse_magnitude_with_crossfade = abs(
            processed_samples[..., idx1] - processed_samples[..., idx0]
        )
        assert seam_impulse_magnitude_with_crossfade < 0.11 * amplitude  # smooth

        assert processed_samples.shape == processed_samples_without_crossfade.shape

    def test_replace_mode_repeats_with_crossfading_at_end(self):
        # Test that it doesn't crash if the end of the array is in the middle of
        # the crossfade of the last part (in replace mode)
        augment = RepeatPart(mode="replace", crossfade_duration=0.124, p=1.0)
        part_num_samples = int(0.29 * 44100)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": part_num_samples,
            "repeats": 5,
            "part_start_index": int(0.22 * 44100),
        }
        augment.freeze_parameters()

        sample_rate = 44100
        amplitude = 0.3
        samples = get_chirp_test(sample_rate, duration=2) * amplitude
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert processed_samples.shape == samples.shape

    def test_replace_mode_many_repeats_with_crossfading(self):
        # Test that it doesn't fade in some of the original audio at the end when there
        # enough repetitions to fill the whole array when in replace mode.
        augment = RepeatPart(mode="replace", crossfade_duration=0.124, p=1.0)
        part_num_samples = int(0.27 * 44100)
        augment.parameters = {
            "should_apply": True,
            "part_num_samples": part_num_samples,
            "repeats": 20,
            "part_start_index": int(0.22 * 44100),
        }
        augment.freeze_parameters()

        sample_rate = 44100
        amplitude = 0.3
        samples = get_chirp_test(sample_rate, duration=2) * amplitude
        processed_samples = augment(samples=samples, sample_rate=sample_rate)
        assert processed_samples.shape == samples.shape

        # The very last part should not have any crossfade into the original signal,
        # where there is only high-freq audio.
        assert_high_frequency_energy_absence(
            processed_samples[int(1.9 * 44100) :], sample_rate
        )

    def test_invalid_parameters(self):
        with pytest.raises(ValueError):
            RepeatPart(crossfade_duration=-0.0001)
        with pytest.raises(ValueError):
            RepeatPart(crossfade_duration=0.0001)
        with pytest.raises(ValueError):
            RepeatPart(min_part_duration=0.0001)
        with pytest.raises(ValueError):
            RepeatPart(min_repeats=0, max_repeats=1)
        with pytest.raises(ValueError):
            RepeatPart(min_repeats=2, max_repeats=1)
        with pytest.raises(ValueError):
            RepeatPart(min_part_duration=0.5, max_part_duration=0.2)
        with pytest.raises(ValueError):
            RepeatPart(
                min_part_duration=0.01, crossfade_duration=0.1
            )
        with pytest.raises(ValueError):
            RepeatPart(mode="append")

    def test_warn_too_short_input(self):
        augmenter = RepeatPart(
            mode="replace",
            min_part_duration=0.1,
            max_part_duration=0.1,
            crossfade_duration=0.0,
            p=1.0,
        )
        samples = np.random.uniform(low=-0.5, high=0.5, size=(100,)).astype(np.float32)
        with pytest.warns(
            UserWarning,
            match=(
                "The input sound is not long enough for applying the RepeatPart"
                " transform with the current parameters."
            ),
        ):
            processed_samples = augmenter(samples, sample_rate=10_000)

        assert_array_equal(processed_samples, samples)

    def test_very_short_crossfade_mask_pair(self):
        fade_in, fade_out = get_crossfade_mask_pair(2)
        assert_array_almost_equal(fade_in, np.array([0.0, 1.0], dtype=np.float32))
        assert_array_almost_equal(fade_out, np.array([1.0, 0.0], dtype=np.float32))



import warnings

import librosa
import numpy as np


def load_sound_file(file_path, sample_rate, mono=True, resample_type="auto"):
    """
    Load an audio file as a floating point time series. Audio will be automatically
    resampled to the given sample rate.

    :param file_path: str or Path instance that points to a sound file
    :param sample_rate: If not None, resample to this sample rate
    :param mono: If True, mix any multichannel data down to mono, and return a 1D array
    :param resample_type: "auto" means use "kaiser_fast" when upsampling and "kaiser_best" when
        downsampling
    """
    file_path = str(file_path)
    samples, actual_sample_rate = librosa.load(
        str(file_path), sr=None, mono=mono, dtype=np.float32
    )

    if sample_rate is not None and actual_sample_rate != sample_rate:
        if resample_type == "auto":
            if librosa.__version__.startswith("0.8."):
                resample_type = (
                    "kaiser_fast" if actual_sample_rate < sample_rate else "kaiser_best"
                )
            else:
                resample_type = "soxr_hq"
        samples = librosa.resample(
            samples,
            orig_sr=actual_sample_rate,
            target_sr=sample_rate,
            res_type=resample_type,
        )
        warnings.warn(
            "{} had to be resampled from {} Hz to {} Hz. This hurt execution time.".format(
                str(file_path), actual_sample_rate, sample_rate
            )
        )

    actual_sample_rate = actual_sample_rate if sample_rate is None else sample_rate

    if mono:
        assert len(samples.shape) == 1
    return samples, actual_sample_rate
